import torch
from joblib import Parallel, delayed
from asr.utils.data_utils import undo_padding
from asr.utils.edit_distance import wer_summary, wer_details_for_batch
from asr.data.dataio.dataio import merge_char, split_word
from asr.data.dataio.wer import print_wer_summary, print_alignments


class MetricStats:
    """A default class for storing and summarizing arbitrary metrics.

    More complex metrics can be created by sub-classing this class.

    Arguments
    ---------
    metric : function
        The function to use to compute the relevant metric. Should take
        at least two arguments (predictions and targets) and can
        optionally take the relative lengths of either or both arguments.
        Not usually used in sub-classes.
    batch_eval: bool
        When True it feeds the evaluation metric with the batched input.
        When False and n_jobs=1, it performs metric evaluation one-by-one
        in a sequential way. When False and n_jobs>1, the evaluation
        runs in parallel over the different inputs using joblib.
    n_jobs : int
        The number of jobs to use for computing the metric. If this is
        more than one, every sample is processed individually, otherwise
        the whole batch is passed at once.
    """

    def __init__(self, metric, n_jobs=1, batch_eval=True):
        self.metric = metric
        self.n_jobs = n_jobs
        self.batch_eval = batch_eval
        self.clear()

    def clear(self):
        """Creates empty container for storage, removing existing stats."""
        self.scores = []
        self.ids = []
        self.summary = {}

    def append(self, ids, *args, **kwargs):
        """Store a particular set of metric scores.

        Arguments
        ---------
        ids : list
            List of ids corresponding to utterances.
        *args, **kwargs
            Arguments to pass to the metric function.
        """
        self.ids.extend(ids)

        # Batch evaluation
        if self.batch_eval:
            scores = self.metric(*args, **kwargs).detach()

        else:
            if "predict" not in kwargs or "target" not in kwargs:
                raise ValueError(
                    "Must pass 'predict' and 'target' as kwargs if batch_eval=False"
                )
            if self.n_jobs == 1:
                # Sequence evaluation (loop over inputs)
                scores = sequence_evaluation(metric=self.metric, **kwargs)
            else:
                # Multiprocess evaluation
                scores = multiprocess_evaluation(
                    metric=self.metric, n_jobs=self.n_jobs, **kwargs
                )

        self.scores.extend(scores)

    def summarize(self, field=None):
        """Summarize the metric scores, returning relevant stats.

        Arguments
        ---------
        field : str
            If provided, only returns selected statistic. If not,
            returns all computed statistics.

        Returns
        -------
        float or dict
            Returns a float if ``field`` is provided, otherwise
            returns a dictionary containing all computed stats.
        """
        min_index = torch.argmin(torch.tensor(self.scores))
        max_index = torch.argmax(torch.tensor(self.scores))
        self.summary = {
            "average": float(sum(self.scores) / len(self.scores)),
            "min_score": float(self.scores[min_index]),
            "min_id": self.ids[min_index],
            "max_score": float(self.scores[max_index]),
            "max_id": self.ids[max_index],
        }

        if field is not None:
            return self.summary[field]
        else:
            return self.summary

    def write_stats(self, filestream, verbose=False):
        """Write all relevant statistics to file.

        Arguments
        ---------
        filestream : file-like object
            A stream for the stats to be written to.
        verbose : bool
            Whether to also print the stats to stdout.
        """
        if not self.summary:
            self.summarize()

        message = f"Average score: {self.summary['average']}\n"
        message += f"Min error: {self.summary['min_score']} "
        message += f"id: {self.summary['min_id']}\n"
        message += f"Max error: {self.summary['max_score']} "
        message += f"id: {self.summary['max_id']}\n"

        filestream.write(message)
        if verbose:
            print(message)


def multiprocess_evaluation(metric, predict, target, lengths=None, n_jobs=8):
    """Runs metric evaluation if parallel over multiple jobs."""
    if lengths is not None:
        lengths = (lengths * predict.size(1)).round().int().cpu()
        predict = [p[:length].cpu() for p, length in zip(predict, lengths)]
        target = [t[:length].cpu() for t, length in zip(target, lengths)]

    while True:
        try:
            scores = Parallel(n_jobs=n_jobs, timeout=30)(
                delayed(metric)(p, t) for p, t in zip(predict, target)
            )
            break
        except Exception as e:
            print(e)
            print("Evaluation timeout...... (will try again)")

    return scores


def sequence_evaluation(metric, predict, target, lengths=None):
    """Runs metric evaluation sequentially over the inputs."""
    if lengths is not None:
        lengths = (lengths * predict.size(1)).round().int().cpu()
        predict = [p[:length].cpu() for p, length in zip(predict, lengths)]
        target = [t[:length].cpu() for t, length in zip(target, lengths)]

    scores = []
    for p, t in zip(predict, target):
        score = metric(p, t)
        scores.append(score)
    return scores


class ErrorRateStats(MetricStats):
    """A class for tracking error rates (e.g., WER, PER).

    Arguments
    ---------
    merge_tokens : bool
        Whether to merge the successive tokens (used for e.g.,
        creating words out of character tokens).
    split_tokens : bool
        Whether to split tokens (used for e.g. creating
        characters out of word tokens).
    space_token : str
        The character to use for boundaries. Used with ``merge_tokens``
        this represents character to split on after merge.
        Used with ``split_tokens`` the sequence is joined with
        this token in between, and then the whole sequence is split.
    """

    def __init__(self, merge_tokens=False, split_tokens=False, space_token="_"):
        self.clear()
        self.merge_tokens = merge_tokens
        self.split_tokens = split_tokens
        self.space_token = space_token

    def append(
        self,
        ids,
        predict,
        target,
        predict_len=None,
        target_len=None,
        ind2lab=None,
    ):
        """Add stats to the relevant containers.

        * See MetricStats.append()

        Arguments
        ---------
        ids : list
            List of ids corresponding to utterances.
        predict : torch.tensor
            A predicted output, for comparison with the target output
        target : torch.tensor
            The correct reference output, for comparison with the prediction.
        predict_len : torch.tensor
            The predictions relative lengths, used to undo padding if
            there is padding present in the predictions.
        target_len : torch.tensor
            The target outputs' relative lengths, used to undo padding if
            there is padding present in the target.
        ind2lab : callable
            Callable that maps from indices to labels, operating on batches,
            for writing alignments.
        """
        self.ids.extend(ids)

        if predict_len is not None:
            predict = undo_padding(predict, predict_len)

        if target_len is not None:
            target = undo_padding(target, target_len)

        if ind2lab is not None:
            predict = ind2lab(predict)
            target = ind2lab(target)

        if self.merge_tokens:
            predict = merge_char(predict, space=self.space_token)
            target = merge_char(target, space=self.space_token)

        if self.split_tokens:
            predict = split_word(predict, space=self.space_token)
            target = split_word(target, space=self.space_token)

        scores = wer_details_for_batch(ids, target, predict, True)

        self.scores.extend(scores)

    def summarize(self, field=None):
        """Summarize the error_rate and return relevant statistics.

        * See MetricStats.summarize()
        """
        self.summary = wer_summary(self.scores)

        # Add additional, more generic key
        self.summary["error_rate"] = self.summary["WER"]

        if field is not None:
            return self.summary[field]
        else:
            return self.summary

    def write_stats(self, filestream):
        """Write all relevant info (e.g., error rate alignments) to file.
        * See MetricStats.write_stats()
        """
        if not self.summary:
            self.summarize()

        print_wer_summary(self.summary, filestream)
        print_alignments(self.scores, filestream)



import subprocess
import yaml
import logging
import time
import os

from e2eAIOK.common.utils import *
from e2eAIOK.SDA.modeladvisor.BaseModelAdvisor import BaseModelAdvisor

class MiniGoAdvisor(BaseModelAdvisor):
    def __init__(self, dataset_meta_path, train_path, eval_path, settings):
        super().__init__(dataset_meta_path, train_path, eval_path, settings)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')
        self.train_path = train_path
        self.test_path = eval_path
        self.dataset_meta_path = dataset_meta_path
        self.data_path = settings['data_path']
        self.checkpoint_dir = f"{settings['data_path']}/checkpoints/mlperf07"
        self.target=f"{settings['data_path']}/target/target.minigo"

    def update_metrics(self):
        with open("result/metric.txt", "r") as f:
            self.winrate = float(f.read())
        metrics = []
        metrics.append({'name': 'winrate', 'value': self.winrate})
        metrics.append({'name': 'training_time', 'value': self.training_time})
        self.params['model_parameter']['metrics'] = metrics
        return self.params['model_parameter']['metrics']

    def initialize_model_parameter(self, assignments=None):
        config = {}
        tuned_parameters = {}
        if assignments:
            tuned_parameters['train_batch_size'] = assignments[
                'train_batch_size']
        else:
            tuned_parameters['train_batch_size'] = 4096
        config['tuned_parameters'] = tuned_parameters
        self.params['model_parameter'] = config
        self.params['model_saved_path'] = os.path.join(
            self.params['save_path'], get_hash_string(str(tuned_parameters)))
        return config

    def generate_sigopt_yaml(self, file='minigo_sigopt.yaml'):
        config = {}
        config['project'] = self.params['model_parameter']['project']
        config['experiment'] = self.params['model_parameter']['experiment']
        parameters = [
        {'name':'train_batch_size','grid':[512,1024,2048,4096,8192],'type':'int'}
        ]
        config['parameters'] = parameters
        config['metrics'] = [{
            'name': 'winrate',
            'objective': 'maximize',
        }, {
            'name': 'training_time',
            'strategy': 'optimize',
            'objective': 'minimize',
        }]
        config['observation_budget'] = self.params['observation_budget']

        saved_path = os.path.join(self.params['save_path'], file)
        with open(saved_path, 'w') as f:
            yaml.dump(config, f)
        return config

    def train_model(self, args):
        start_time = time.time()
        self.dist_launch(args)
        self.training_time = time.time() - start_time
        metrics = self.update_metrics()
        model_path = args['model_saved_path']
        return self.training_time, model_path, metrics

    def dist_launch(self, args):
        hosts = args['hosts']
        rootnode = args['rootnode']
        os.chdir('modelzoo/minigo')
        pre_cmd=f"sed -i '/--train_batch_size=/ s/=.*/={args['model_parameter']['tuned_parameters']['train_batch_size']}/' ml_perf/flags/19/train.flags"
        process = subprocess.Popen(pre_cmd, shell=True)
        process.wait()
        cmd = f"HOSTLIST={','.join(hosts)} ROOTNODE={rootnode} checkpoint_dir={self.checkpoint_dir} target={self.target} ml_perf/scripts/run_minigo.sh 19"
        self.logger.info(f'training launch command: {cmd}')
        process = subprocess.Popen(cmd, shell=True)
        process.wait()
        clean_cmd = f"pkill -f ml_perf"
        process = subprocess.Popen(clean_cmd, shell=True)
        process.wait()
        os.chdir('../../')



import sqlite3
import os
from e2eAIOK.utils.hydromodel import *


class HydroDB:
    """
    This class is used to create one instance for hydro database
    """
    def __init__(self, hydrodb_path):
        self.hydrodb_path = hydrodb_path
        if not os.path.exists(self.hydrodb_path):
            self.__init_db()
        else:
            self.conn = sqlite3.connect(self.hydrodb_path)
            self.cur = self.conn.cursor()

    def __del__(self):
        self.conn.close()

    def __init_db(self):
        self.conn = sqlite3.connect(self.hydrodb_path)
        self.cur = self.conn.cursor()
        self.cur.execute(
            '''CREATE TABLE models (learner_id text, hydro_model json)''')
        self.conn.commit()

    def print_all(self):
        for row in self.cur.execute("SELECT * FROM models"):
            print(row)

    def get_model_by_id(self, learner_id):
        """
        Get in database model info by learner_id

        Parameters
        ----------
        learner_id : str
            Learning_id is identical to task

        Returns
        -------
        hydro_model: HydroModel
        """
        hydro_model = None
        for row in self.cur.execute(
                "SELECT hydro_model FROM models WHERE learner_id = '%s'" %
                learner_id):
            hydro_model = HydroModel(None, row)
        return hydro_model

    def insert_model_by_id(self, learner_id, hydro_model):
        """
        Insert new or update history model by learner_id

        Parameters
        ----------
        learner_id : str
            Learning_id is identical to task
        hydro_model: HydroModel
        """
        # if learner_id exsits, update original one
        # or insert new one
        do_update = False
        for row in self.cur.execute(
                "SELECT * FROM models WHERE learner_id = '%s'" % learner_id):
            do_update = True
        if do_update:
            self.cur.execute(
                "UPDATE models SET hydro_model = '%s' WHERE learner_id = '%s'"
                % (hydro_model.to_json(), learner_id))
        else:
            self.cur.execute("INSERT INTO models VALUES ('%s','%s')" %
                             (learner_id, hydro_model.to_json()))
        self.conn.commit()

    def clear(self):
        """
        Remove e2eaiok.db file
        """
        os.remove(self.hydrodb_path)


# Todo: Add common cv model features into this module in the future
import torch
from e2eAIOK.common.trainer.model_builder import ModelBuilder

class ModelBuilderCV(ModelBuilder):
    def __init__(self, cfg, model=None):
        super().__init__(cfg, model)



import torch.nn as nn


class EmbeddingBase(nn.Embedding):
    def __init__(self, num_embeddings, embedding_dim, blank_id=None):
        super().__init__(num_embeddings, embedding_dim, padding_idx=blank_id)

    def forward(self, x):
        raise NotImplementedError



# coding=utf-8
# Copyright (c) 2022, Intel. and its affiliates.
# Copyright (c) 2019, Facebook, Inc. and its affiliates. All Rights Reserved
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tokenization classes."""

from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import logging
import os
import unicodedata
# from io import open


logger = logging.getLogger(__name__)

PRETRAINED_VOCAB_ARCHIVE_MAP = {
    'bert-base-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-uncased-vocab.txt",
    'bert-large-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt",
    'bert-base-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-cased-vocab.txt",
    'bert-large-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt",
    'spanbert-base-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-cased-vocab.txt",
    'spanbert-large-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt"
}

PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP = {
    'bert-base-uncased': 512,
    'bert-large-uncased': 512,
    'bert-base-cased': 512,
    'bert-large-cased': 512,
    'spanbert-base-cased': 512,
    'spanbert-large-cased': 512
}
VOCAB_NAME = 'vocab.txt'


def load_vocab(vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = collections.OrderedDict()
    index = 0
    with open(vocab_file, "r", encoding="utf-8") as reader:
        while True:
            token = reader.readline()
            if not token:
                break
            token = token.strip()
            vocab[token] = index
            index += 1
    return vocab


def whitespace_tokenize(text):
    """Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip()
    if not text:
        return []
    tokens = text.split()
    return tokens


class BertTokenizer(object):
    """Runs end-to-end tokenization: punctuation splitting + wordpiece"""

    def __init__(self, vocab_file, do_lower_case=True, max_len=None, do_basic_tokenize=True,
                 never_split=("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]")):
        """Constructs a BertTokenizer.

        Args:
          vocab_file: Path to a one-wordpiece-per-line vocabulary file
          do_lower_case: Whether to lower case the input
                         Only has an effect when do_wordpiece_only=False
          do_basic_tokenize: Whether to do basic tokenization before wordpiece.
          max_len: An artificial maximum length to truncate tokenized sequences to;
                         Effective maximum length is always the minimum of this
                         value (if specified) and the underlying BERT model's
                         sequence length.
          never_split: List of tokens which will never be split during tokenization.
                         Only has an effect when do_wordpiece_only=False
        """
        if not os.path.isfile(vocab_file):
            raise ValueError(
                "Can't find a vocabulary file at path '{}'. To load the vocabulary from a Google pretrained "
                "model use `tokenizer = BertTokenizer.from_pretrained(PRETRAINED_MODEL_NAME)`".format(vocab_file))
        self.vocab = load_vocab(vocab_file)
        self.ids_to_tokens = collections.OrderedDict(
            [(ids, tok) for tok, ids in self.vocab.items()])
        self.do_basic_tokenize = do_basic_tokenize
        if do_basic_tokenize:
            self.basic_tokenizer = BasicTokenizer(do_lower_case=do_lower_case,
                                                  never_split=never_split)
        self.wordpiece_tokenizer = WordpieceTokenizer(vocab=self.vocab)
        self.max_len = max_len if max_len is not None else int(1e12)

    def tokenize(self, text):
        split_tokens = []
        if self.do_basic_tokenize:
            for token in self.basic_tokenizer.tokenize(text):
                for sub_token in self.wordpiece_tokenizer.tokenize(token):
                    split_tokens.append(sub_token)
        else:
            split_tokens = self.wordpiece_tokenizer.tokenize(text)
        return split_tokens

    def convert_tokens_to_ids(self, tokens):
        """Converts a sequence of tokens into ids using the vocab."""
        ids = []
        for token in tokens:
            ids.append(self.vocab[token])
        if len(ids) > self.max_len:
            logger.warning(
                "Token indices sequence length is longer than the specified maximum "
                " sequence length for this BERT model ({} > {}). Running this"
                " sequence through BERT will result in indexing errors".format(len(ids), self.max_len)
            )
        return ids

    def convert_ids_to_tokens(self, ids):
        """Converts a sequence of ids in wordpiece tokens using the vocab."""
        tokens = []
        for i in ids:
            tokens.append(self.ids_to_tokens[i])
        return tokens

    def save_vocabulary(self, vocab_path):
        """Save the tokenizer vocabulary to a directory or file."""
        index = 0
        if os.path.isdir(vocab_path):
            vocab_file = os.path.join(vocab_path, VOCAB_NAME)
        with open(vocab_file, "w", encoding="utf-8") as writer:
            for token, token_index in sorted(self.vocab.items(), key=lambda kv: kv[1]):
                if index != token_index:
                    logger.warning("Saving vocabulary to {}: vocabulary indices are not consecutive."
                                   " Please check that the vocabulary is not corrupted!".format(vocab_file))
                    index = token_index
                writer.write(token + u'\n')
                index += 1
        return vocab_file

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, cache_dir=None, *inputs, **kwargs):
        """
        Instantiate a PreTrainedBertModel from a pre-trained model file.
        Download and cache the pre-trained model file if needed.
        """
        if pretrained_model_name_or_path in PRETRAINED_VOCAB_ARCHIVE_MAP:
            vocab_file = PRETRAINED_VOCAB_ARCHIVE_MAP[pretrained_model_name_or_path]
            if '-cased' in pretrained_model_name_or_path and kwargs.get('do_lower_case', True):
                logger.warning("The pre-trained model you are loading is a cased model but you have not set "
                               "`do_lower_case` to False. We are setting `do_lower_case=False` for you but "
                               "you may want to check this behavior.")
                kwargs['do_lower_case'] = False
            elif '-cased' not in pretrained_model_name_or_path and not kwargs.get('do_lower_case', True):
                logger.warning("The pre-trained model you are loading is an uncased model but you have set "
                               "`do_lower_case` to False. We are setting `do_lower_case=True` for you "
                               "but you may want to check this behavior.")
                kwargs['do_lower_case'] = True
        else:
            vocab_file = pretrained_model_name_or_path
        if os.path.isdir(vocab_file):
            vocab_file = os.path.join(vocab_file, VOCAB_NAME)
        # redirect to the cache, if necessary
        try:
            resolved_vocab_file = vocab_file
        except EnvironmentError:
            logger.error(
                "Model name was not found in model name list. "
                "We assumed 'model_name' was a path or url but couldn't find any file "
                "associated to this path or url.")
            return None
        logger.info("loading vocabulary file")
        if pretrained_model_name_or_path in PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP:
            # if we're using a pretrained model, ensure the tokenizer wont index sequences longer
            # than the number of positional embeddings
            max_len = PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP[pretrained_model_name_or_path]
            kwargs['max_len'] = min(kwargs.get('max_len', int(1e12)), max_len)
        # Instantiate tokenizer.
        tokenizer = cls(resolved_vocab_file, *inputs, **kwargs)
        return tokenizer


class BasicTokenizer(object):
    """Runs basic tokenization (punctuation splitting, lower casing, etc.)."""

    def __init__(self,
                 do_lower_case=True,
                 never_split=("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]")):
        """Constructs a BasicTokenizer.

        Args:
          do_lower_case: Whether to lower case the input.
        """
        self.do_lower_case = do_lower_case
        self.never_split = never_split

    def tokenize(self, text):
        """Tokenizes a piece of text."""
        text = self._clean_text(text)
        # This was added on November 1st, 2018 for the multilingual and Chinese
        # models. This is also applied to the English models now, but it doesn't
        # matter since the English models were not trained on any Chinese data
        # and generally don't have any Chinese data in them (there are Chinese
        # characters in the vocabulary because Wikipedia does have some Chinese
        # words in the English Wikipedia.).
        text = self._tokenize_chinese_chars(text)
        orig_tokens = whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            if self.do_lower_case and token not in self.never_split:
                token = token.lower()
                token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token))

        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    def _run_strip_accents(self, text):
        """Strips accents from a piece of text."""
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    def _run_split_on_punc(self, text):
        """Splits punctuation on a piece of text."""
        if text in self.never_split:
            return [text]
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if _is_punctuation(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _tokenize_chinese_chars(self, text):
        """Adds whitespace around any CJK character."""
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, cp):
        """Checks whether CP is the codepoint of a CJK character."""
        # This defines a "chinese character" as anything in the CJK Unicode block:
        #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
        #
        # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        # despite its name. The modern Korean Hangul alphabet is a different block,
        # as is Japanese Hiragana and Katakana. Those alphabets are used to write
        # space-separated words, so they are not treated specially and handled
        # like the all of the other languages.
        if ((cp >= 0x4E00 and cp <= 0x9FFF) or  #
                (cp >= 0x3400 and cp <= 0x4DBF) or  #
                (cp >= 0x20000 and cp <= 0x2A6DF) or  #
                (cp >= 0x2A700 and cp <= 0x2B73F) or  #
                (cp >= 0x2B740 and cp <= 0x2B81F) or  #
                (cp >= 0x2B820 and cp <= 0x2CEAF) or
                (cp >= 0xF900 and cp <= 0xFAFF) or  #
                (cp >= 0x2F800 and cp <= 0x2FA1F)):  #
            return True

        return False

    def _clean_text(self, text):
        """Performs invalid character removal and whitespace cleanup on text."""
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xfffd or _is_control(char):
                continue
            if _is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)


class WordpieceTokenizer(object):
    """Runs WordPiece tokenization."""

    def __init__(self, vocab, unk_token="[UNK]", max_input_chars_per_word=100):
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word

    def tokenize(self, text):
        """Tokenizes a piece of text into its word pieces.

        This uses a greedy longest-match-first algorithm to perform tokenization
        using the given vocabulary.

        For example:
          input = "unaffable"
          output = ["un", "##aff", "##able"]

        Args:
          text: A single token or whitespace separated tokens. This should have
            already been passed through `BasicTokenizer`.

        Returns:
          A list of wordpiece tokens.
        """

        output_tokens = []
        for token in whitespace_tokenize(text):
            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens


def _is_whitespace(char):
    """Checks whether `chars` is a whitespace character."""
    # \t, \n, and \r are technically contorl characters but we treat them
    # as whitespace since they are generally considered as such.
    if char == " " or char == "\t" or char == "\n" or char == "\r":
        return True
    cat = unicodedata.category(char)
    if cat == "Zs":
        return True
    return False


def _is_control(char):
    """Checks whether `chars` is a control character."""
    # These are technically control characters but we count them as whitespace
    # characters.
    if char == "\t" or char == "\n" or char == "\r":
        return False
    cat = unicodedata.category(char)
    if cat.startswith("C"):
        return True
    return False


def _is_punctuation(char):
    """Checks whether `chars` is a punctuation character."""
    cp = ord(char)
    # We treat all non-letter/number ASCII as punctuation.
    # Characters such as "^", "$", and "`" are not in the Unicode
    # Punctuation class but we treat them as punctuation anyways, for
    # consistency.
    if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
            (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False



import os
import torch
import logging
import numpy as np
import pickle
import csv
import time
import torchaudio
import json
import re
logger = logging.getLogger(__name__)


def load_data_json(json_path, replacements={}):
    """Loads JSON and recursively formats string values.

    Arguments
    ----------
    json_path : str
        Path to CSV file.
    replacements : dict
        (Optional dict), This is used to recursively format all string values in the data.

    Returns
    -------
    dict
        JSON data with replacements applied.
    """
    with open(json_path, "r") as f:
        out_json = json.load(f)
    _recursive_format(out_json, replacements)
    return out_json


def _recursive_format(data, replacements):
    if isinstance(data, dict):
        for key, item in data.items():
            if isinstance(item, dict) or isinstance(item, list):
                _recursive_format(item, replacements)
            elif isinstance(item, str):
                data[key] = item.format_map(replacements)
            # If not dict, list or str, do nothing
    if isinstance(data, list):
        for i, item in enumerate(data):
            if isinstance(item, dict) or isinstance(item, list):
                _recursive_format(item, replacements)
            elif isinstance(item, str):
                data[i] = item.format_map(replacements)
            # If not dict, list or str, do nothing


def load_data_csv(csv_path, replacements={}):
    """Loads CSV and formats string values.

    CSV must have an 'ID' field.
    If there is a field called duration, it is interpreted as a float.
    The rest of the fields are left as they are (legacy _format and _opts fields
    are not used to load the data in any special way).

    Bash-like string replacements with $to_replace are supported.

    Arguments
    ----------
    csv_path : str
        Path to CSV file.
    replacements : dict
        (Optional dict), This is used to recursively format all string values in the data.

    Returns
    -------
    dict
        CSV data with replacements applied.
    """

    with open(csv_path, newline="") as csvfile:
        result = {}
        reader = csv.DictReader(csvfile, skipinitialspace=True)
        variable_finder = re.compile(r"\$([\w.]+)")
        for row in reader:
            # ID:
            try:
                data_id = row["ID"]
                del row["ID"]  # This is used as a key in result, instead.
            except KeyError:
                raise KeyError(
                    "CSV has to have an 'ID' field, with unique ids"
                    " for all data points"
                )
            if data_id in result:
                raise ValueError(f"Duplicate id: {data_id}")
            # Replacements:
            for key, value in row.items():
                try:
                    row[key] = variable_finder.sub(
                        lambda match: str(replacements[match[1]]), value
                    )
                except KeyError:
                    raise KeyError(
                        f"The item {value} requires replacements "
                        "which were not supplied."
                    )
            # Duration:
            if "duration" in row:
                row["duration"] = float(row["duration"])
            result[data_id] = row
    return result


def read_audio(waveforms_obj):
    """General audio loading, based on a custom notation.

    Arguments
    ----------
    waveforms_obj : str, dict
        Audio reading annotation, see above for format.

    Returns
    -------
    torch.Tensor
        Audio tensor with shape: (samples, ).
    """
    if isinstance(waveforms_obj, str):
        audio, _ = torchaudio.load(waveforms_obj)
        return audio.transpose(0, 1).squeeze(1)

    path = waveforms_obj["file"]
    start = waveforms_obj.get("start", 0)
    # Default stop to start -> if not specified, num_frames becomes 0,
    # which is the torchaudio default
    stop = waveforms_obj.get("stop", start)
    num_frames = stop - start
    audio, fs = torchaudio.load(path, num_frames=num_frames, frame_offset=start)
    audio = audio.transpose(0, 1)
    return audio.squeeze(1)


def length_to_mask(length, max_len=None, dtype=None, device=None):
    """Creates a binary mask for each sequence.

    Reference: https://discuss.pytorch.org/t/how-to-generate-variable-length-mask/23397/3

    Arguments
    ---------
    length : torch.LongTensor
        Containing the length of each sequence in the batch. Must be 1D.
    max_len : int
        Max length for the mask, also the size of the second dimension.
    dtype : torch.dtype, default: None
        The dtype of the generated mask.
    device: torch.device, default: None
        The device to put the mask variable.

    Returns
    -------
    mask : tensor
        The binary mask.
    """
    assert len(length.shape) == 1

    if max_len is None:
        max_len = length.max().long().item()  # using arange to generate mask
    mask = torch.arange(
        max_len, device=length.device, dtype=length.dtype
    ).expand(len(length), max_len) < length.unsqueeze(1)

    if dtype is None:
        dtype = length.dtype

    if device is None:
        device = length.device

    mask = torch.as_tensor(mask, dtype=dtype, device=device)
    return mask


def save_pkl(obj, file):
    """Save an object in pkl format.

    Arguments
    ---------
    obj : object
        Object to save in pkl format
    file : str
        Path to the output file
    sampling_rate : int
        Sampling rate of the audio file, TODO: this is not used?
    """
    with open(file, "wb") as f:
        pickle.dump(obj, f)


def load_pkl(file):
    """Loads a pkl file.

    Arguments
    ---------
    file : str
        Path to the input pkl file.

    Returns
    -------
    The loaded object.
    """

    # Deals with the situation where two processes are trying
    # to access the same label dictionary by creating a lock
    count = 100
    while count > 0:
        if os.path.isfile(file + ".lock"):
            time.sleep(1)
            count -= 1
        else:
            break

    try:
        with open(file + ".lock", "w"):
            pass
        with open(file, "rb") as f:
            return pickle.load(f)
    finally:
        if os.path.isfile(file + ".lock"):
            os.remove(file + ".lock")


def merge_char(sequences, space="_"):
    """Merge characters sequences into word sequences.

    Arguments
    ---------
    sequences : list
        Each item contains a list, and this list contains a character sequence.
    space : string
        The token represents space. Default: _

    Returns
    -------
    The list contains word sequences for each sentence.

    Example
    -------
    >>> sequences = [["a", "b", "_", "c", "_", "d", "e"], ["e", "f", "g", "_", "h", "i"]]
    >>> results = merge_char(sequences)
    >>> results
    [['ab', 'c', 'de'], ['efg', 'hi']]
    """
    results = []
    for seq in sequences:
        words = "".join(seq).split(space)
        results.append(words)
    return results


def merge_csvs(data_folder, csv_lst, merged_csv):
    """Merging several csv files into one file.

    Arguments
    ---------
    data_folder : string
        The folder to store csv files to be merged and after merging.
    csv_lst : list
        Filenames of csv file to be merged.
    merged_csv : string
        The filename to write the merged csv file.

    Example
    -------
    >>> merge_csvs("tests/samples/annotation/",
    ... ["speech.csv", "speech.csv"],
    ... "test_csv_merge.csv")
    """
    write_path = os.path.join(data_folder, merged_csv)
    if os.path.isfile(write_path):
        logger.info("Skipping merging. Completed in previous run.")
    with open(os.path.join(data_folder, csv_lst[0])) as f:
        header = f.readline()
    lines = []
    for csv_file in csv_lst:
        with open(os.path.join(data_folder, csv_file)) as f:
            for i, line in enumerate(f):
                if i == 0:
                    # Checking header
                    if line != header:
                        raise ValueError(
                            "Different header for " f"{csv_lst[0]} and {csv}."
                        )
                    continue
                lines.append(line)
    with open(write_path, "w") as f:
        f.write(header)
        for line in lines:
            f.write(line)
    logger.info(f"{write_path} is created.")


def split_word(sequences, space="_"):
    """Split word sequences into character sequences.

    Arguments
    ---------
    sequences : list
        Each item contains a list, and this list contains a words sequence.
    space : string
        The token represents space. Default: _

    Returns
    -------
    The list contains word sequences for each sentence.

    Example
    -------
    >>> sequences = [['ab', 'c', 'de'], ['efg', 'hi']]
    >>> results = split_word(sequences)
    >>> results
    [['a', 'b', '_', 'c', '_', 'd', 'e'], ['e', 'f', 'g', '_', 'h', 'i']]
    """
    results = []
    for seq in sequences:
        chars = list(space.join(seq))
        results.append(chars)
    return results



import time
import sigopt

from e2eAIOK.DeNas.search.BaseSearchEngine import BaseSearchEngine
from e2eAIOK.common.utils import timeout_input

class MOSigoptSearchEngine(BaseSearchEngine):
    def __init__(self, params=None, super_net=None, search_space=None):
        super().__init__(params,super_net,search_space)
        self.conn = None
        self.vis_dict = {}
        self.best_struct = ()

    def _get_sigopt_suggestion(self, experiment):
        num_tried = 0
        while True:
            try:
                return self.conn.experiments(experiment.id).suggestions().create()
            except Exception as e:
                num_tried += 1
                self.logger.error("""Met exception when connecting to sigopt,
                    will do retry in 5 secs, err msg is: {}""".format(e))
                if num_tried >= 30:
                    n = timeout_input(
                        """Retried connection for 30 times, do you still
                        want to continue?(n for exit)""",
                        default='y',
                        timeout=10)
                    if n != 'y':
                        return None
                    num_tried = 0
                time.sleep(5)

    def _set_sigopt_observation(self, experiment, suggestion_id, metrics):
        num_tried = 0
        while True:
            try:
                return self.conn.experiments(experiment.id).observations().create(
                    suggestion=suggestion_id,
                    values=metrics
                )
            except Exception as e:
                num_tried += 1
                self.logger.error("""Met exception when connecting to sigopt,
                    will do retry in 5 secs, err msg is: {}""".format(e))
                if num_tried >= 30:
                    n = timeout_input(
                        """Retried connection for 30 times, do you still
                        want to continue?(n for exit)""",
                        default='y',
                        timeout=10)
                    if n != 'y':
                        return None
                    num_tried = 0
                time.sleep(5)

    def _set_illegal_observation(self, experiment, suggestion_id):
        num_tried = 0
        while True:
            try:
                return self.conn.experiments(experiment.id).observations().create(
                    suggestion=suggestion_id,
                    failed=True
                )
            except Exception as e:
                num_tried += 1
                self.logger.error("""Met exception when connecting to sigopt,
                    will do retry in 5 secs, err msg is: {}""".format(e))
                if num_tried >= 30:
                    n = timeout_input(
                        """Retried connection for 30 times, do you still
                        want to continue?(n for exit)""",
                        default='y',
                        timeout=10)
                    if n != 'y':
                        return None
                    num_tried = 0
                time.sleep(5)
    '''
    Unified API for SigoptSearchEngine
    '''
    def search(self):
        self.conn = sigopt.Connection()
        if self.params.domain == "bert":
            experiment = self.conn.experiments().create(
                name= 'bert denas',
                project='denas',
                type="offline",
                observation_budget=1000,
                metrics=[dict(name='Score', objective='maximize'),dict(name='Latency', objective='minimize')],
                parameters=[
                    dict(name="LAYER_NUM", type="int", bounds=dict(min=self.params.cfg["SEARCH_SPACE"]['LAYER_NUM']['bounds']['min'], max=self.params.cfg["SEARCH_SPACE"]['LAYER_NUM']['bounds']['max'])),
                    dict(name="HEAD_NUM", type="int", bounds=dict(min=self.params.cfg["SEARCH_SPACE"]['HEAD_NUM']['bounds']['min'], max=self.params.cfg["SEARCH_SPACE"]['HEAD_NUM']['bounds']['max']-1)),
                    dict(name="HIDDEN_SIZE", type="int", bounds=dict(min=self.params.cfg["SEARCH_SPACE"]['HIDDEN_SIZE']['bounds']['min']/self.params.cfg["SEARCH_SPACE"]['HIDDEN_SIZE']['bounds']['step'], max=self.params.cfg["SEARCH_SPACE"]['HIDDEN_SIZE']['bounds']['max']/self.params.cfg["SEARCH_SPACE"]['HIDDEN_SIZE']['bounds']['step']-1)),
                    dict(name="INTERMEDIATE_SIZE", type="int", bounds=dict(min=self.params.cfg["SEARCH_SPACE"]['INTERMEDIATE_SIZE']['bounds']['min']/self.params.cfg["SEARCH_SPACE"]['INTERMEDIATE_SIZE']['bounds']['step'], max=self.params.cfg["SEARCH_SPACE"]['INTERMEDIATE_SIZE']['bounds']['max']/self.params.cfg["SEARCH_SPACE"]['INTERMEDIATE_SIZE']['bounds']['step']-1)),
                ],
            )
            for epoch in range(experiment.observation_budget):
                suggestion = self._get_sigopt_suggestion(experiment)
                cand = (
                    suggestion.assignments['LAYER_NUM'], 
                    suggestion.assignments['HEAD_NUM'], 
                    64*suggestion.assignments['HEAD_NUM'], 
                    suggestion.assignments['HIDDEN_SIZE']*self.params.cfg["SEARCH_SPACE"]['HIDDEN_SIZE']['bounds']['step'], 
                    suggestion.assignments['INTERMEDIATE_SIZE']*self.params.cfg["SEARCH_SPACE"]['INTERMEDIATE_SIZE']['bounds']['step'],
                )
                if not self.cand_islegal(cand):
                    self._set_illegal_observation(experiment, suggestion.id)
                    continue
                if not self.cand_islegal_latency(cand):
                    self._set_illegal_observation(experiment, suggestion.id)
                    continue
                nas_score, score, latency = self.cand_evaluate(cand)
                self.logger.info('epoch = {} structure = {} nas_score = {} params = {}'.format(epoch, cand, self.vis_dict[cand]['score'], self.vis_dict[cand]['params']))
                metrics = []
                metrics.append({'name': 'Score', 'value': score.item()})
                metrics.append({'name': 'Latency', 'value': latency})
                self._set_sigopt_observation(experiment, suggestion.id, metrics)
            best_assignments = self.conn.experiments(experiment.id).best_assignments().fetch().data[0].assignments
            self.best_struct = (best_assignments['LAYER_NUM'],best_assignments['HEAD_NUM'],64*best_assignments['HEAD_NUM'],best_assignments['HIDDEN_SIZE']*self.params.cfg["SEARCH_SPACE"]['HIDDEN_SIZE']['bounds']['step'],best_assignments['INTERMEDIATE_SIZE']*self.params.cfg["SEARCH_SPACE"]['INTERMEDIATE_SIZE']['bounds']['step'])
        elif self.params.domain == "vit":
            experiment = self.conn.experiments().create(
                name= 'vit denas',
                project='denas',
                type="offline",
                observation_budget=1000,
                metrics=[dict(name='Score', objective='maximize'),dict(name='Latency', objective='minimize')],
                conditionals=[dict(name="DEPTH",values=["12","13","14","15","16"])],
                parameters=[
                    dict(name="MLP_RATIO_0", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_1", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_2", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_3", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_4", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_5", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_6", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_7", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_8", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_9", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_10", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_11", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_12", type="double", grid=[*self.search_space['mlp_ratio']], conditions=dict(DEPTH=["16","15","14","13"])),
                    dict(name="MLP_RATIO_13", type="double", grid=[*self.search_space['mlp_ratio']], conditions=dict(DEPTH=["16","15","14"])),
                    dict(name="MLP_RATIO_14", type="double", grid=[*self.search_space['mlp_ratio']], conditions=dict(DEPTH=["16","15"])),
                    dict(name="MLP_RATIO_15", type="double", grid=[*self.search_space['mlp_ratio']], conditions=dict(DEPTH=["16"])),
                    dict(name="NUM_HEADS_0", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_1", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_2", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_3", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_4", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_5", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_6", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_7", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_8", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_9", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_10", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_11", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_12", type="int", grid=[*self.search_space['num_heads']], conditions=dict(DEPTH=["16","15","14","13"])),
                    dict(name="NUM_HEADS_13", type="int", grid=[*self.search_space['num_heads']], conditions=dict(DEPTH=["16","15","14"])),
                    dict(name="NUM_HEADS_14", type="int", grid=[*self.search_space['num_heads']], conditions=dict(DEPTH=["16","15"])),
                    dict(name="NUM_HEADS_15", type="int", grid=[*self.search_space['num_heads']], conditions=dict(DEPTH=["16"])),
                    dict(name="EMBED_DIM", type="int", grid=[*self.search_space['embed_dim']]),
                ],
            )
            for epoch in range(experiment.observation_budget):
                suggestion = self._get_sigopt_suggestion(experiment)
                cand_tuple = list()
                depth = int(suggestion.assignments['DEPTH'])
                cand_tuple.append(depth)
                for i in range(depth):
                    mlp_ratio_name = f"MLP_RATIO_{i}"
                    cand_tuple.append(float(suggestion.assignments[mlp_ratio_name]))
                for i in range(depth):
                    num_heads_name = f"NUM_HEADS_{i}"
                    cand_tuple.append(int(suggestion.assignments[num_heads_name]))
                cand_tuple.append(int(suggestion.assignments['EMBED_DIM']))
                cand = tuple(cand_tuple)
                if not self.cand_islegal(cand):
                    self._set_illegal_observation(experiment, suggestion.id)
                    continue
                if not self.cand_islegal_latency(cand):
                    self._set_illegal_observation(experiment, suggestion.id)
                    continue
                nas_score, score, latency = self.cand_evaluate(cand)
                self.logger.info('epoch = {} structure = {} nas_score = {} params = {}'.format(epoch, cand, self.vis_dict[cand]['score'], self.vis_dict[cand]['params']))
                metrics = []
                metrics.append({'name': 'Score', 'value': score.item()})
                metrics.append({'name': 'Latency', 'value': latency})
                self._set_sigopt_observation(experiment, suggestion.id, metrics)
            best_assignments = self.conn.experiments(experiment.id).best_assignments().fetch().data[0].assignments  
            depth = int(best_assignments['DEPTH'])
            best_tuple = list()
            best_tuple.append(depth)
            for i in range(depth):
                mlp_ratio_name = f"MLP_RATIO_{i}"
                best_tuple.append(float(best_assignments[mlp_ratio_name]))
            for i in range(depth):
                num_heads_name = f"NUM_HEADS_{i}"
                best_tuple.append(int(best_assignments[num_heads_name]))
            best_tuple.append(int(best_assignments['EMBED_DIM']))
            self.best_struct = tuple(best_tuple)
        elif self.params.domain == "asr":
            experiment = self.conn.experiments().create(
                name= 'asr denas',
                project='denas',
                type="offline",
                observation_budget=1000,
                metrics=[dict(name='Score', objective='maximize'),dict(name='Latency', objective='minimize')],
                conditionals=[dict(name="DEPTH",values=["9","10","11","12"])],
                parameters=[
                    dict(name="MLP_RATIO_0", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_1", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_2", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_3", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_4", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_5", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_6", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_7", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_8", type="double", grid=[*self.search_space['mlp_ratio']]),
                    dict(name="MLP_RATIO_9", type="double", grid=[*self.search_space['mlp_ratio']], conditions=dict(DEPTH=["12","11","10"])),
                    dict(name="MLP_RATIO_10", type="double", grid=[*self.search_space['mlp_ratio']], conditions=dict(DEPTH=["12","11"])),
                    dict(name="MLP_RATIO_11", type="double", grid=[*self.search_space['mlp_ratio']], conditions=dict(DEPTH=["12"])),
                    dict(name="NUM_HEADS_0", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_1", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_2", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_3", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_4", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_5", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_6", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_7", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_8", type="int", grid=[*self.search_space['num_heads']]),
                    dict(name="NUM_HEADS_9", type="int", grid=[*self.search_space['num_heads']], conditions=dict(DEPTH=["12","11","10"])),
                    dict(name="NUM_HEADS_10", type="int", grid=[*self.search_space['num_heads']], conditions=dict(DEPTH=["12","11"])),
                    dict(name="NUM_HEADS_11", type="int", grid=[*self.search_space['num_heads']], conditions=dict(DEPTH=["12"])),
                    dict(name="EMBED_DIM", type="int", grid=[*self.search_space['embed_dim']]),
                ],
            )
            for epoch in range(experiment.observation_budget):
                suggestion = self._get_sigopt_suggestion(experiment)
                cand_tuple = list()
                depth = int(suggestion.assignments['DEPTH'])
                cand_tuple.append(depth)
                for i in range(depth):
                    mlp_ratio_name = f"MLP_RATIO_{i}"
                    cand_tuple.append(float(suggestion.assignments[mlp_ratio_name]))
                for i in range(depth):
                    num_heads_name = f"NUM_HEADS_{i}"
                    cand_tuple.append(int(suggestion.assignments[num_heads_name]))
                cand_tuple.append(int(suggestion.assignments['EMBED_DIM']))
                cand = tuple(cand_tuple)
                if not self.cand_islegal(cand):
                    self._set_illegal_observation(experiment, suggestion.id)
                    continue
                nas_score, score, latency = self.cand_evaluate(cand)
                self.logger.info('epoch = {} nas_score = {} cand = {}'.format(epoch, nas_score, cand))
                metrics = []
                metrics.append({'name': 'Score', 'value': score.item()})
                metrics.append({'name': 'Latency', 'value': latency})
                self._set_sigopt_observation(experiment, suggestion.id, metrics)
            best_assignments = self.conn.experiments(experiment.id).best_assignments().fetch().data[0].assignments  
            depth = int(best_assignments['DEPTH'])
            best_tuple = list()
            best_tuple.append(depth)
            for i in range(depth):
                mlp_ratio_name = f"MLP_RATIO_{i}"
                best_tuple.append(float(best_assignments[mlp_ratio_name]))
            for i in range(depth):
                num_heads_name = f"NUM_HEADS_{i}"
                best_tuple.append(int(best_assignments[num_heads_name]))
            best_tuple.append(int(best_assignments['EMBED_DIM']))
            self.best_struct = tuple(best_tuple)
        else:
            raise RuntimeError(f"Domain {self.params.domain} is not supported")
        with open("best_model_structure.txt", 'w') as f:
            f.write(str(self.best_struct))

    '''
    Unified API to get best searched structure
    '''
    def get_best_structures(self):
        self.logger.info('best structure {} nas_score {} params {}'.format(self.best_struct, self.vis_dict[self.best_struct]['score'], self.vis_dict[self.best_struct]['params']))
        return self.best_struct



__all__ = ['BaseModelAdvisor', 'BERTAdvisor', 'DIENAdvisor', 'DLRMAdvisor', 'MiniGoAdvisor', 'RegisteredAdvisor', 'ResNetAdvisor', 'RNNTAdvisor', 'TestAdvisor', 'TwitterRecSysAdvisor', 'UPMAdvisor', 'WnDAdvisor']


import torch
import logging

from asr.utils import checkpoints

logger = logging.getLogger(__name__)


@checkpoints.register_checkpoint_hooks
class NoamScheduler:
    """The is an implementation of the transformer's learning rate scheduler with warmup.
    Reference: https://arxiv.org/abs/1706.03762

    Note: this scheduler anneals the lr at each update of the model's weight,
    and n_steps must be saved for restarting.

    Arguments
    ---------
    lr_initial : float
        Initial learning rate (i.e. the lr used at epoch 0).
    n_warmup_steps : int
        numer of warm-up steps
    model_size : int
        size of transformer embed_dim. It is used to scale the maximum learning rate value reached
        by the scheduler. It is divided by model_size ** (0.5).
        If not specified the maximum learning rate value is instead multiplied by warmup_steps ** (0.5).
    """

    def __init__(self, lr_initial, n_warmup_steps, model_size=None):
        self.lr_initial = lr_initial
        self.n_warmup_steps = n_warmup_steps
        self.current_lr = lr_initial
        self.losses = []
        self.n_steps = 0
        self.normalize = n_warmup_steps ** 0.5
        if model_size is not None:
            self.normalize = model_size ** (-0.5)

    def __call__(self, opt):
        """
        Arguments
        ---------
        opt : optimizer
            The optimizer to update using this scheduler.

        Returns
        -------
        current_lr : float
            The learning rate before the update.
        lr : float
            The learning rate after the update.
        """
        self.n_steps += 1

        current_lr = opt.param_groups[0]["lr"]

        lr = self.lr_initial * self._get_lr_scale()

        # Changing the learning rate within the optimizer
        for param_group in opt.param_groups:
            param_group["lr"] = lr

        self.current_lr = current_lr
        return current_lr, lr

    def _get_lr_scale(self):
        n_steps, n_warmup_steps = self.n_steps, self.n_warmup_steps
        return self.normalize * min(
            n_steps ** (-0.5), n_steps * n_warmup_steps ** (-1.5)
        )

    @checkpoints.mark_as_saver
    def save(self, path):
        """Saves the current metrics on the specified path."""
        data = {"losses": self.losses, "n_steps": self.n_steps}
        torch.save(data, path)

    @checkpoints.mark_as_loader
    def load(self, path, end_of_epoch=False, device=None):
        """Loads the needed information."""
        del end_of_epoch  # Unused in this class
        del device
        data = torch.load(path)
        self.losses = data["losses"]
        self.n_steps = data["n_steps"]



import torch


class SpecAugment(torch.nn.Module):
    """An implementation of the SpecAugment algorithm.

    Reference:
        https://arxiv.org/abs/1904.08779

    Arguments
    ---------
    time_warp : bool
        Whether applying time warping.
    time_warp_window : int
        Time warp window.
    time_warp_mode : str
        Interpolation mode for time warping (default "bicubic").
    freq_mask : bool1
        Whether applying freq mask.
    freq_mask_width : int or tuple
        Freq mask width range.
    n_freq_mask : int
        Number of freq mask.
    time_mask : int
        Whether applying time mask.
    time_mask_width : int or tuple
        Time mask width range.
    n_time_mask : int
        Number of time mask.
    replace_with_zero : bool
        If True, replace masked value with 0, else replace masked value with mean of the input tensor.
    """

    def __init__(
        self,
        time_warp=True,
        time_warp_window=5,
        time_warp_mode="bicubic",
        freq_mask=True,
        freq_mask_width=(0, 20),
        n_freq_mask=2,
        time_mask=True,
        time_mask_width=(0, 100),
        n_time_mask=2,
        replace_with_zero=True,
    ):
        super().__init__()
        assert (
            time_warp or freq_mask or time_mask
        ), "at least one of time_warp, time_mask, or freq_mask should be applied"

        self.apply_time_warp = time_warp
        self.time_warp_window = time_warp_window
        self.time_warp_mode = time_warp_mode

        self.freq_mask = freq_mask
        if isinstance(freq_mask_width, int):
            freq_mask_width = (0, freq_mask_width)
        self.freq_mask_width = freq_mask_width
        self.n_freq_mask = n_freq_mask

        self.time_mask = time_mask
        if isinstance(time_mask_width, int):
            time_mask_width = (0, time_mask_width)
        self.time_mask_width = time_mask_width
        self.n_time_mask = n_time_mask

        self.replace_with_zero = replace_with_zero

    def forward(self, x):
        """Takes in input a tensors and returns an augmented one."""
        if self.apply_time_warp:
            x = self.time_warp(x)
        if self.freq_mask:
            x = self.mask_along_axis(x, dim=2)
        if self.time_mask:
            x = self.mask_along_axis(x, dim=1)
        return x

    def time_warp(self, x):
        """Time warping with torch.nn.functional.interpolate"""
        original_size = x.shape
        window = self.time_warp_window

        # 2d interpolation requires 4D or higher dimension tensors
        # x: (Batch, Time, Freq) -> (Batch, 1, Time, Freq)
        if x.dim() == 3:
            x = x.unsqueeze(1)

        time = x.shape[2]
        if time - window <= window:
            return x.view(*original_size)

        # compute center and corresponding window
        c = torch.randint(window, time - window, (1,))[0]
        w = torch.randint(c - window, c + window, (1,))[0] + 1

        left = torch.nn.functional.interpolate(
            x[:, :, :c],
            (w, x.shape[3]),
            mode=self.time_warp_mode,
            align_corners=True,
        )
        right = torch.nn.functional.interpolate(
            x[:, :, c:],
            (time - w, x.shape[3]),
            mode=self.time_warp_mode,
            align_corners=True,
        )

        x[:, :, :w] = left
        x[:, :, w:] = right

        return x.view(*original_size)

    def mask_along_axis(self, x, dim):
        """Mask along time or frequency axis.

        Arguments
        ---------
        x : tensor
            Input tensor.
        dim : int
            Corresponding dimension to mask.
        """
        original_size = x.shape
        if x.dim() == 4:
            x = x.view(-1, x.shape[2], x.shape[3])

        batch, time, fea = x.shape

        if dim == 1:
            D = time
            n_mask = self.n_time_mask
            width_range = self.time_mask_width
        else:
            D = fea
            n_mask = self.n_freq_mask
            width_range = self.freq_mask_width

        mask_len = torch.randint(
            width_range[0], width_range[1], (batch, n_mask), device=x.device
        ).unsqueeze(2)

        mask_pos = torch.randint(
            0, max(1, D - mask_len.max()), (batch, n_mask), device=x.device
        ).unsqueeze(2)

        # compute masks
        arange = torch.arange(D, device=x.device).view(1, 1, -1)
        mask = (mask_pos <= arange) * (arange < (mask_pos + mask_len))
        mask = mask.any(dim=1)

        if dim == 1:
            mask = mask.unsqueeze(2)
        else:
            mask = mask.unsqueeze(1)

        if self.replace_with_zero:
            val = 0.0
        else:
            val = x.mean()

        x = x.masked_fill_(mask, val)
        return x.view(*original_size)



import json
import logging
import sys
import time
from collections import OrderedDict
import yaml
from sigopt import Connection

from e2eAIOK.common.utils import *
from e2eAIOK.utils.hydroconfig import *

class BaseModelAdvisor:
    """Model Advisor Base, Model Advisor is used to create w/wo sigopt
    parameter advise based on model type

    Attributes
    ----------
    conn
        sigopt connection
    experiment
        current experiment which is either created by this run or
        fetched by history experiment id
    params : dict
        params include dataset_path, save_path, global_configs,
        model_parameters, passed by arguments or e2eaiok-defaults.conf
    assignment_model_tracker : dict
        a tracker map of assigned_parameters and its corresponding
        model path

    """
    def __init__(self, dataset_meta_path, train_path, eval_path, settings):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')
        self.interative = settings["interative"]

        self.conn = None
        self.experiment = None
        self.params = init_advisor_params()
        self.params['dataset_meta_path'] = dataset_meta_path
        self.params['train_dataset_path'] = train_path
        self.params['eval_dataset_path'] = eval_path
        dest_saved_path = os.path.join(self.params['save_path'],
                                       settings['model_name'])
        self.params['save_path'] = mkdir(dest_saved_path)
        self.assignment_model_tracker = {}
        self.params.update(settings)

    def initialize_sigopt(self, experiment_id=None):
        """
        This method is used to prepare sigopt including call
        generate_sigopt_yaml and setup_sigopt_connection

        Parameters
        ----------
        experiment_id: str
          history sigopt experiment_id for this task, when this is not
          None, we can use this id to fetch previous experiment

        Returns
        -------
        sigopt_experiment_id: str
          sigopt experiment id for this task
        """
        # 1. create sigopt yaml
        sigopt_config = self.generate_sigopt_yaml()
        yaml.dump(sigopt_config, sys.stdout)
        n = timeout_input("Please confirm with sigopt parameters?(n for exit)",
                          default='y', interactive=self.interative)
        if n != 'y':
            exit()
        self.params['model_parameter'] = sigopt_config
        # 2. create sigopt connection
        self.conn, self.experiment = self._setup_sigopt_connection(
            sigopt_config, experiment_id)
        return self.experiment.id

    def register(self, info):
        for k, v in info.items():
            if k == "score_metrics" and 'metrics' in dir(self):
                if not isinstance(v, list) or not (len(v) > 0 and isinstance(v[0], tuple)):
                    raise ValueError("[Register advisor] scores metrics should be a pair, ex: ('mean_accuracy', 'maximize')")
                self.metrics = v
            if k == "experiment_name" and 'experiment_name' in dir(self):
                self.experiment_name = v
            if k == "tune_training_time" and 'training_time_as_metrics' in dir(self):
                self.training_time_as_metrics = True
            if k == "hyper_parameters" and 'parameters' in dir(self):
                if not isinstance(v, dict):
                    raise ValueError("""
        [Register advisor] hyper_parameters should be a dict, ex: 
        {'max_depth':11, 'learning_rate':float(0.9), 'min_split_loss':float(7)}""")
                self.parameters = v
            if k == "sigopt_config" and 'sigopt_config' in dir(self):
                if not isinstance(v, list) or not (len(v) > 0 and isinstance(v[0], dict)):
                    raise ValueError("""
        [Register advisor] sigopt_config should be a list, ex: 
        [{
            'name': 'learning_rate',
            'bounds': {
                'min': 0.0,
                'max': 1.0
            },
            'type': 'double'
        }]""")
                self.sigopt_config = v
                self.parameters = dict((i["name"], None) for i in v)
            if k == "execute_cmd" and "execute_cmd_base" in dir(self):
                self.execute_cmd_base = v
            if k == "result_file_name" and "result_file_name" in dir(self):
                self.result_file_name = v
            if k == "observation_budget" and "observation_budget" in dir(self):
                self.observation_budget = v

    def _setup_sigopt_connection(self, sigopt_config, experiment_id=None):
        name = sigopt_config["experiment"]
        parameters = sigopt_config["parameters"]
        metrics = sigopt_config["metrics"]
        observation_budget = sigopt_config["observation_budget"]
        project = sigopt_config["project"]

        num_tried = 0
        while True:
            try:
                conn = Connection()
                if experiment_id:
                    experiment = conn.experiments(experiment_id).fetch()
                else:
                    experiment = conn.experiments().create(
                        name=name,
                        parameters=parameters,
                        metrics=metrics,
                        observation_budget=observation_budget,
                        project=project)
                return conn, experiment
            except Exception as e:
                num_tried += 1
                print(
                    """[WARNING] Met exception when connecting to sigopt,
                    will do retry in 5 secs, err msg is: """
                )
                print(e)
                if num_tried >= 30:
                    n = timeout_input(
                        """Retried connection for 30 times, do you still
                        want to continue?(n for exit)""",
                        default='y',
                        timeout=10)
                    if n != 'y':
                        return None, None
                    num_tried = 0
                time.sleep(5)

    def _assignments_to_string(self, assignments):
        tmp = assignments
        if not isinstance(assignments, dict):
            tmp = json.loads(json.dumps(assignments))
        tmp = str(OrderedDict(sorted(tmp.items())))
        return tmp

    def _assign_sigopt_suggestion(self, sigopt_assignments):
        config = self.initialize_model_parameter(sigopt_assignments)
        self.params['model_saved_path'] = os.path.join(
            self.params['save_path'], get_hash_string(str(config)))

    def record_assignmet(self, assignments, model_path):
        """
        This method is used to record latest assignment with its trained model path

        Parameters
        ----------
        assignment: Assignment
          sigopt assignment of model parameters
        model_path: str
          trained model saved path
        """
        if not assignments:
            return
        tmp = self._assignments_to_string(assignments)
        self.assignment_model_tracker[tmp] = model_path

    def find_model_by_assignment(self, assignments):
        """
        This method is used to find model saved path by corresponding assignment

        Parameters
        ----------
        assignment: Assignment
          sigopt assignment of model parameters

        Returns
        -------
        model_path: str
          trained model saved path
        """
        tmp = self._assignments_to_string(assignments)
        if tmp in self.assignment_model_tracker:
            return self.assignment_model_tracker[tmp]
        else:
            print(f"can't find {tmp} in {self.assignment_model_tracker}")
            return None

    def get_metrics_from_best_assignments(self, best_assignments):
        """
        This method is used to get metrics from BestAssignments

        Parameters
        ----------
        best_assignments: BestAssignments
          sigopt returned best_assignment

        Returns
        -------
        metrics: list
        """

        metrics = []
        for item in best_assignments.values:
            metrics.append({"name": item.name, "value": item.value})
        return metrics

    def get_best_parameters_from_best_assignments(self, best_assignments):
        """
        This method is used to get model parameter assignment from BestAssignments

        Parameters
        ----------
        best_assignments: BestAssignments
          sigopt returned best_assignment

        Returns
        -------
        assignment: Assignment
        """
        saved_path = os.path.join(self.params['save_path'],
                                  "best_parameters.yaml")
        with open(saved_path, 'w') as f:
            yaml.dump(best_assignments.assignments, f)
        return best_assignments.assignments

    def _launch_train_with_sigopt(self):
        budget = self.experiment.observation_budget
        self.logger.info(f'Experiment budget: {budget}')
        for _ in range(budget):
            suggestion = self.conn.experiments(
                self.experiment.id).suggestions().create()
            assignments = suggestion.assignments
            self._assign_sigopt_suggestion(assignments)
            mkdir_or_backup_then_mkdir(self.params['model_saved_path'])
            model_path = ""
            metrics = []
            training_time, model_path, metrics = self.train_model(self.params)
            self.logger.info(
                f'Received sigopt suggestion, assignment is \
                    {self._assignments_to_string(assignments)}'
            )
            self.record_assignmet(assignments, model_path)
            self.logger.info(
                f'Training completed with sigopt suggestion, \
                    metrics is {metrics}'
            )

            num_tried = 0
            while True:
                try:
                    self.conn.experiments(self.experiment.id).observations().create(
                        suggestion=suggestion.id, values=metrics)
                    self.experiment = self.conn.experiments(self.experiment.id).fetch()
                    break
                except Exception as e:
                    num_tried += 1
                    print(
                        f'[WARNING] Met exception when connecting to \
                            sigopt, will do retry in 5 secs, err msg is:'
                    )
                    print(e)
                    if num_tried >= 30:
                        n = timeout_input(
                            "Retried connection for 30 times, do you \
                                still want to continue?(n for exit)",
                            default='y',
                            timeout=10)
                        if n != 'y':
                            return model_path, metrics, assignments
                        num_tried = 0
                    time.sleep(5)

        self.logger.info(
            f"Training with sigopt is completed! \
                https://app.sigopt.com/experiment/{self.experiment.id} "
        )
        all_best_assignments = self.conn.experiments(
            self.experiment.id).best_assignments().fetch()
        if len(all_best_assignments.data) == 0:
            self.logger.error(
                "No assignments for satisfied model, you may increase \
                    observation budget or modify metric value"
            )
            return model_path, metrics, assignments
        else:
            model_path = None
            best_assignments = self.get_best_parameters_from_best_assignments(
                all_best_assignments.data[0])
            model_path = self.find_model_by_assignment(best_assignments)
            metrics = self.get_metrics_from_best_assignments(
                all_best_assignments.data[0])
            return model_path, metrics, best_assignments

    def _launch_train_without_sigopt(self):
        mkdir_or_backup_then_mkdir(self.params['model_saved_path'])
        training_time, model_path, metrics = self.train_model(self.params)
        self.logger.info(
            f'Training completed based in sigopt suggestion, took {training_time} secs'
        )
        return model_path, metrics, self.params['model_parameter'][
            "tuned_parameters"]

    def launch_train(self):
        """
        This method will launch train w/wo sigopt according to config

        Returns
        -------
        model_path: str
            best trained model saved path
        metrics: list
            A list of this best trained model infomation
        parameters: dict
            A dict of best model parameters for this training
        """
        if self.experiment is not None:
            return self._launch_train_with_sigopt()
        else:
            return self._launch_train_without_sigopt()

    # Abstract method, Any advisor should have an implementation of below

    def generate_sigopt_yaml(self):
        """
        generate parameter range based on model type and passed-in params for sigopt

        Returns
        -------
        sigopt_parameters: dict
          sigopt parameter including ranged model parameter and metrics based on model type
        """
        raise NotImplementedError("generate_sigopt_yaml is abstract.")

    def initialize_model_parameter(self, assignments=None):
        """
        generate parameters based on model type for wo sigopt case

        Parameters
        ----------
        assignments: dict
          optional user defined or sigopt suggested parameters

        Returns
        -------
        model_parameters: dict
          model parameter based on model type
        """
        # expect return parameter(dict), this is for no_sigopt path
        raise NotImplementedError("initialize_model_parameter is abstract.")

    def train_model(self, args):
        """
        launch train with passed-in model parameters

        Parameters
        ----------
        args: dict
          global parameters like pnn, host; model_parameter from sigopt
          suggestion or model pre-defined

        Returns
        -------
        training_time: float
        model_path: str
        metrics: list
        """
        raise NotImplementedError("train_model is abstract.")

    def update_metrics(self):
        """
        update return metrics based on training result, normally
        training script should have a stdout output or a result.yaml
        file, and this method will pass this file and return metrics

        Returns
        -------
        metrics: list
          metrics from training output, maybe AUC, MAP, training_time, etc
        """
        # expect return updated metrics(list of dict)
        raise NotImplementedError("update_metrics is abstract.")



import torch
import torch.nn as nn
import torch.nn.functional as F

from module.Linear_base import LinearBase


class Linear(LinearBase):
    """
    Computes a linear transformation y = wx + b.
    """

    def __init__(
        self,
        n_neurons,
        input_size=None,
        bias=True,
    ):
        super().__init__(input_size, n_neurons, bias=bias)

    def forward(self, x):
        return F.linear(x, self.weight, self.bias)



class DeSearchSpace:
    def __init__():
        pass

class DeMainNet:
    def __init__():
        pass

    @classmethod
    def create_netblock_list_from_str(cls, struct_str, no_create=False, **kwargs):
        pass


import torch.nn as nn


class LinearBase(nn.Linear):
    def __init__(self, in_dim, out_dim, bias=True):
        super().__init__(in_dim, out_dim, bias=bias)

    def forward(self, x):
        raise NotImplementedError



import inspect
from dataclasses import dataclass
from asr.utils.depgraph import DependencyGraph


@dataclass
class StaticItem:
    """Data class that represents a static item.

    Static items are in-memory items so they don't need to be computed
    dynamically.
    """

    key: str


class DynamicItem:
    """Essentially represents a data transformation function.

    A DynamicItem takes some arguments and computes its value dynamically when
    called. A straight-forward use-case is to load something from disk
    dynamically; take the path and provide the loaded data.

    Instances of this class are often created implicitly via the
    @takes and @provides decorators or otherwise from specifying the taken and
    provided arguments and the function.

    A counterpart is the GeneratorDynamicItem, which should be used for
    generator functions.

    Arguments
    ---------
    takes : list
        The keys of the items that this needs to compute its output.
    func : callable
        The function that is used to compute the output.
    provides : list
        The keys that this provides.
    """

    def __init__(self, takes=[], func=None, provides=[]):
        self.takes = takes
        self.func = func
        self.provides = provides

    def __call__(self, *args):
        return self.func(*args)

    # The next methods are more about supporting GeneratorDynamicItems
    def next_takes(self):
        """The next argkeys to provide to this, when called."""
        # Regular function DynamicItems always just need the same set of args
        return self.takes

    def next_provides(self):
        """The next keys that this provides, when called."""
        # Regular function DynamicItems always just provide the same set of keys
        return self.provides

    def provided_in_order(self):
        """Assuming that this may need to be called multiple times; which keys
        does it provide at that call. Returns a list, with len equal to the
        number of times that this may be called."""
        # Regular function DynamicItems are only called once:
        return [self.provides]

    def reset(self):
        """Signals that this will not be called any more times on this pipeline
        call."""
        # Regular function DynamicItems don't need special resets.
        pass


class GeneratorDynamicItem(DynamicItem):
    """Essentially represents a multi-step data transformation.

    This is the generator function counterpart for DynamicItem (which should be
    used for regular functions).

    A GeneratorDynamicItem first takes some arguments and then uses those in
    multiple steps to incrementally compute some values when called.

    A typical use-case is a pipeline of transformations on data: e.g. taking in
    text as a string, and first a tokenized version, and then on the second
    call providing an integer-encoded version. This can be used even though the
    integer-encoder needs to be trained on the first outputs.

    The main benefit is to be able to define the pipeline in a clear function,
    even if parts of the pipeline depend on others for their initialization.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Doesn't generate electricity, only stores the currently active
        # generator:
        self.current_generator = None
        self.num_provided_items = 0

    def __call__(self, *args):
        if self.num_provided_items == len(self.provides):
            raise RuntimeError("DynamicItemPipeline called too many times!")
        if not self.current_generator:
            self.current_generator = self.func(*args)
        # NOTE: Not supporting sending new values to the pipeline.
        out = next(self.current_generator)
        self.num_provided_items += 1
        return out

    def next_takes(self):
        """The next argkeys to provide to this, when called."""
        if not self.current_generator:
            return self.takes
        else:
            return []

    def next_provides(self):
        """The next keys that this provides, when called."""
        keys = self.provides[self.num_provided_items]
        # Support multiple yielded values like:
        # @yields("wav_read", ["left_ch", "right_ch"])
        if isinstance(keys, str):
            return [keys]
        else:
            return keys

    def provided_in_order(self):
        """Assuming that this may need to be called multiple times; which keys
        does it provide at that call. Returns a list, with len equal to the
        number of times that this may be called."""
        in_order = []
        for keys in self.provides:
            # Support multiple yielded values like:
            # @provides("wav_read", ["left_ch", "right_ch"])
            if isinstance(keys, str):
                in_order.append([keys])
            else:
                in_order.append(keys)
        return in_order

    def reset(self):
        """Signals that this will not be called any more times on this pipeline
        call."""
        if self.current_generator is not None:
            self.current_generator.close()
        self.current_generator = None
        self.num_provided_items = 0


def takes(*argkeys):
    """Decorator which makes a DynamicItem and specifies its argkeys.

    If the wrapped object is a generator function (has a yield statement),
    Creates a GeneratorDynamicItem. If the object is already a DynamicItem,
    just specifies the argkeys for that. Otherwise creates a new regular
    DynamicItem, with argkeys specified.

    The args are always passed to the function at the start. Generators could
    support sending new arguments, but for such use cases, simply create a new
    dynamic item. The GeneratorDynamicItem class is meant for pipelines which
    take in an input and transform it in multiple ways, where the intermediate
    representations may be needed for e.g. fitting a BPE segmenter.
    """

    def decorator(obj):
        """Decorator definition."""
        if isinstance(obj, DynamicItem):
            if obj.takes:
                raise ValueError("Can't overwrite DynamicItem.takes")
            obj.takes = argkeys
            return obj
        elif inspect.isgeneratorfunction(obj):
            return GeneratorDynamicItem(takes=argkeys, func=obj)
        else:
            return DynamicItem(takes=argkeys, func=obj)

    return decorator


takes_decorator = takes  # Just for DataPipeline.add_dynamic_item


def provides(*output_keys):
    """Decorator which makes a DynamicItem and specifies what keys it provides.

    If the wrapped object is a generator function (has a yield statement),
    Creates a GeneratorDynamicItem. If the object is already a DynamicItem,
    just specifies the provided keys for that. Otherwise creates a new regular
    DynamicItem, with provided keys specified.

    NOTE
    ----
    The behavior is slightly different for generators and regular functions, if
    many output keys are specified, e.g. @provides("signal", "mfcc"). Regular
    functions should return a tuple with len equal to len(output_keys), while
    generators should yield the items one by one.
    """

    def decorator(obj):
        """Decorator definition."""
        if isinstance(obj, DynamicItem):
            if obj.provides:
                raise ValueError("Can't overwrite DynamicItem provides-list.")
            obj.provides = output_keys
            return obj
        elif inspect.isgeneratorfunction(obj):
            return GeneratorDynamicItem(func=obj, provides=output_keys)
        else:
            return DynamicItem(func=obj, provides=output_keys)

    return decorator


provides_decorator = provides  # Just for DataPipeline.add_dynamic_item


class DataPipeline:
    """Organises data transformations into a pipeline.
    """

    def __init__(self, static_data_keys, dynamic_items=[], output_keys=[]):
        self.dg = DependencyGraph()
        self._exec_order = None
        self.key_to_node = {}
        self.unaccounted_keys = {}
        self.dynamic_items = []
        self.output_mapping = {}
        self.add_static_keys(static_data_keys)
        self.add_dynamic_items(dynamic_items)
        self.set_output_keys(output_keys)

    def add_static_keys(self, static_keys):
        """Informs the pipeline about static items.

        Static items are the ones provided to __call__ as data.
        """
        for key in static_keys:
            node_id = self.dg.add_node(data=StaticItem(key=key))
            self.key_to_node[key] = node_id

    def add_dynamic_items(self, dynamic_items):
        """Add multiple dynamic items at once."""
        for item in dynamic_items:
            try:
                self.add_dynamic_item(**item)
            except TypeError:
                self.add_dynamic_item(item)

    def add_dynamic_item(self, func, takes=None, provides=None):
        """Adds a dynamic item to the Pipeline.

        Two calling conventions. For DynamicItem objects, just use:
        add_dynamic_item(dynamic_item)
        But otherwise, should use:
        add_dynamic_item(func, takes, provides)

        Arguments
        ---------
        func : callable, DynamicItem
            If a DynamicItem is given, adds that directly. Otherwise a
            DynamicItem is created, and this specifies the callable to use. If
            a generator function is given, then create a GeneratorDynamicItem.
            Otherwise creates a normal DynamicItem.
        takes : list, str
            List of keys. When func is called, each key is resolved to
            either an entry in the data or the output of another dynamic_item.
            The func is then called with these as positional arguments,
            in the same order as specified here.
            A single key can be given as a bare string.
        provides : str, list
            For regular functions, the key or list of keys that it provides.
            If you give a generator function, key or list of keys that it
            yields, in order. Also see the provides decorator.
            A single key can be given as a bare string.
        """
        if isinstance(func, DynamicItem):
            if takes is not None or provides is not None:
                raise ValueError(
                    "If providing a DynamicItem directly, don't "
                    "specify takes or provides"
                )
            else:
                self._add_dynamic_item_object(func)
                return
        if isinstance(takes, str):
            takes = [takes]
        if isinstance(provides, str):
            provides = [provides]
        di = takes_decorator(*takes)(provides_decorator(*provides)(func))
        self._add_dynamic_item_object(di)

    def _add_dynamic_item_object(self, obj):
        """Internally adds the object.

        There is a node in the dependency graph for each call of the
        DynamicItem. Each call may return multiple keys and depend on multiple
        keys. An internal dict maps key to the id of the node that produces it.
        """
        if not obj.provides:
            raise ValueError(
                "Won't add redundant dynamic item which doesn't "
                "provide anything."
            )
        depended = []
        for key in obj.takes:
            # Might not be accounted for, yet:
            if key not in self.key_to_node:
                dependee_keys = self.unaccounted_keys.setdefault(key, [])
                dependee_keys.extend(obj.next_provides())
            else:
                depended.append(self.key_to_node[key])
        for provided in obj.provided_in_order():
            node_id = self.dg.add_node(data=obj)
            for key in provided:
                self.key_to_node[key] = node_id
                # This key may also be unaccounted for, so account for it now:
                if key in self.unaccounted_keys:
                    for dependee_key in self.unaccounted_keys[key]:
                        dependee_node = self.key_to_node[dependee_key]
                        self.dg.add_edge(dependee_node, node_id)
                    del self.unaccounted_keys[key]  # Now accounted for!
            for dep_id in depended:
                self.dg.add_edge(node_id, dep_id)
            # Next call will depend on this call:
            depended = [node_id]
        # Keep a reference to the item in this object, as well:
        self.dynamic_items.append(obj)

    def set_output_keys(self, keys):
        """Use this to change the output keys.

        Also re-evaluates execution order.
        So if you request different outputs, some parts of the
        data pipeline may be skipped.

        Arguments
        ---------
        keys : dict, list, None
            List of keys (str) to produce in output.

            If a dict is given; it is used to map internal keys to output keys.
            From the output_keys dict key:value pairs the key appears outside,
            and value is the internal key.
        """
        self.output_mapping = self._output_keys_to_mapping(keys)
        self._exec_order = None

    @staticmethod
    def _output_keys_to_mapping(keys):
        # Ensure a mapping (accept a list for convenience, too)
        if keys is None:
            output_mapping = {}
        elif isinstance(keys, dict):
            output_mapping = keys
        else:
            output_mapping = {key: key for key in keys}
        return output_mapping

    def compute_outputs(self, data):
        """
        Arguments
        ---------
        data : dict
            Dictionary with data entries by key.

        Returns
        -------
        dict
            With the keys that were set.
        """
        if self._exec_order is None:
            self._prepare_run(data)
        return self._compute(data, self._exec_order, self.output_mapping)

    def compute_specific(self, keys, data):
        """Compute output of specific item, without changing output_keys."""
        output_mapping = self._output_keys_to_mapping(keys)
        order = self.dg.get_evaluation_order(
            selected_keys=self.get_selected_node_ids(keys)
        )
        return self._compute(data, order, output_mapping)

    def _compute(self, data, order, output_mapping):
        if self.unaccounted_keys:
            MSG = "These keys are still unaccounted for in the data pipeline: "
            MSG += ", ".join(self.unaccounted_keys)
            raise RuntimeError(MSG)
        intermediate = {}
        for node_id, edges, item in order:
            if isinstance(item, StaticItem):
                # Static item in data.
                # Just check that key is found.
                try:
                    data[item.key]
                    continue
                except KeyError:
                    raise KeyError(f"Expected key {item.key} in data!")
            # A dynamic item, which we should compute:
            args = [
                data[argkey] if argkey in data else intermediate[argkey]
                for argkey in item.next_takes()
            ]
            # This needs to be called BEFORE the dynamic item is called.
            provided_keys = item.next_provides()
            values = item(*args)  # Call the DynamicItem to produce output
            # If there is just one output value, wrap in a list so that
            # it can be zipped as well:
            if len(provided_keys) == 1:
                values = [values]
            intermediate.update(zip(provided_keys, values))
        for dynamic_item in self.dynamic_items:
            dynamic_item.reset()
        return {
            outkey: data[inkey] if inkey in data else intermediate[inkey]
            for outkey, inkey in output_mapping.items()
        }

    def get_selected_node_ids(self, selected_keys):
        """Translates selected keys to dependency graph keys."""
        return [self.key_to_node[key] for key in selected_keys]

    def __call__(self, data):
        return self.compute_outputs(data)

    def _prepare_run(self, data):
        self._exec_order = list(
            self.dg.get_evaluation_order(
                self.get_selected_node_ids(self.output_mapping.values())
            )
        )



import os
import sys
current_dir = os.path.split(os.path.abspath(__file__))[0]
cls_path = current_dir.rsplit("/", 3)[0]
sys.path.append(cls_path)
import sys
import torch
import logging
import time
import yaml
import sentencepiece as sp
import torch.distributed as dist

from asr.data.dataio.dataloader import get_dataloader, make_dataloader
from asr.data.dataio.dataset import dataio_prepare
from asr.utils.utils import check_gradients, update_average, create_experiment_directory, init_log
from asr.utils.parameter_transfer import load_torch_model, load_spm
from asr.trainer.losses import ctc_loss, kldiv_loss
from asr.trainer.schedulers import NoamScheduler
from asr.data.augment import SpecAugment
from asr.data.features import Fbank
from asr.utils.Accuracy import AccuracyStats
from asr.utils.metric_stats import ErrorRateStats
from e2eAIOK.common.trainer.torch_trainer import TorchTrainer
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist

class ASRTrainer(TorchTrainer):
    def __init__(self, cfg, model, train_dataloader, eval_dataloader, optimizer, criterion, scheduler, metric, tokenizer):
        super(ASRTrainer, self).__init__(cfg, model, train_dataloader, eval_dataloader, optimizer, criterion, scheduler, metric)
        self.tokenizer = tokenizer
    
    def _pre_process(self):
        super()._pre_process()
        load_spm(self.tokenizer, self.cfg.tokenizer_ckpt)
        self.feat_proc = Fbank(**self.cfg["compute_features"])
    
    def _is_early_stop(self, metric):
        return metric <= self.cfg["metric_threshold"]
    
    def _dist_wrapper(self):
        if ext_dist.my_size > 1:
            for name, module in self.model.items():
                if any(p.requires_grad for p in module.parameters()):
                    module = torch.nn.SyncBatchNorm.convert_sync_batchnorm(module)
                    module = ext_dist.DDP(module)
                    self.model[name] = module

    def train_one_epoch(self, epoch):
        augment = SpecAugment(**self.cfg["augmentation"])
        if self.train_dataloader.sampler is not None and hasattr(self.train_dataloader.sampler, "set_epoch"):
            self.train_dataloader.sampler.set_epoch(epoch)
        self.model.train()

        step = 0
        nonfinite_count = 0
        total_step = len(self.train_dataloader)
        avg_train_loss = 0.0
        epoch_start_time = time.time()
        for batch in self.train_dataloader:
            step += 1
            step_start_time = time.time()
            should_step = step % self.cfg["grad_accumulation_factor"] == 0
            wavs, wav_lens = batch.sig
            tokens_bos, _ = batch.tokens_bos
            feats = self.feat_proc(wavs)
            feats = self.model["normalize"](feats, wav_lens, epoch=epoch)
            feats = augment(feats)

            src = self.model["CNN"](feats)
            enc_out, pred = self.model["Transformer"](src, tokens_bos, wav_lens, pad_idx=self.cfg["pad_index"])

            logits = self.model["ctc_lin"](enc_out)
            p_ctc = logits.log_softmax(dim=-1)

            pred = self.model["seq_lin"](pred)
            p_seq = pred.log_softmax(dim=-1)

            ids = batch.id
            tokens_eos, tokens_eos_lens = batch.tokens_eos
            tokens, tokens_lens = batch.tokens

            # loss_seq = kldiv_loss(p_seq, tokens_eos, length=tokens_eos_lens, label_smoothing=self.cfg["label_smoothing"], reduction=self.cfg["loss_reduction"]).sum()
            # loss_ctc = ctc_loss(p_ctc, tokens, wav_lens, tokens_lens, blank_index=self.cfg["blank_index"], reduction=self.cfg["loss_reduction"]).sum()
            loss_seq = self.criterion["seq_loss"](p_seq, tokens_eos, length=tokens_eos_lens, label_smoothing=self.cfg["label_smoothing"], reduction=self.cfg["loss_reduction"]).sum()
            loss_ctc = self.criterion["ctc_loss"](p_ctc, tokens, wav_lens, tokens_lens, blank_index=self.cfg["blank_index"], reduction=self.cfg["loss_reduction"]).sum()

            loss = (self.cfg["ctc_weight"] * loss_ctc + (1 - self.cfg["ctc_weight"]) * loss_seq)
            (loss / self.cfg["grad_accumulation_factor"]).backward()

            if should_step:
                is_loss_finite, nonfinite_count = check_gradients(self.model, loss, self.cfg["max_grad_norm"], nonfinite_count)
                if is_loss_finite:
                    self.optimizer.step()
                self.optimizer.zero_grad()
                self.scheduler(self.optimizer)

            train_loss = loss.detach().cpu()
            avg_train_loss = update_average(train_loss, avg_train_loss, step)
            self.logger.info(f"epoch: {epoch}, step: {step}|{total_step}, time: {(time.time()-step_start_time):.2f}s, loss: {train_loss}, avg_loss: {avg_train_loss:.4f}, lr: {self.scheduler.current_lr}")

        self.logger.info(f"epoch: {epoch}, time: {(time.time()-epoch_start_time):.2f}s, avg_loss: {avg_train_loss:.4f}")

    def evaluate(self, epoch):
        self.metric.clear()
        self.model.eval()
        avg_valid_loss = 0.0
        total_step = len(self.eval_dataloader)
        step = 0
        eval_start_time = time.time()
        with torch.no_grad():
            for batch in self.eval_dataloader:
                step += 1
                step_start_time = time.time()
                wavs, wav_lens = batch.sig
                tokens_bos, _ = batch.tokens_bos
                feats = self.feat_proc(wavs)
                feats = self.model["normalize"](feats, wav_lens, epoch=epoch)

                src = self.model["CNN"](feats)
                enc_out, pred = self.model["Transformer"](src, tokens_bos, wav_lens, pad_idx=self.cfg["pad_index"])

                logits = self.model["ctc_lin"](enc_out)
                p_ctc = logits.log_softmax(dim=-1)

                pred = self.model["seq_lin"](pred)
                p_seq = pred.log_softmax(dim=-1)

                # hyps, _ = searcher(enc_out.detach(), wav_lens)

                ids = batch.id
                tokens_eos, tokens_eos_lens = batch.tokens_eos
                tokens, tokens_lens = batch.tokens

                # loss_seq = kldiv_loss(p_seq, tokens_eos, length=tokens_eos_lens, label_smoothing=self.cfg["label_smoothing"], reduction=self.cfg["loss_reduction"]).sum()
                # loss_ctc = ctc_loss(p_ctc, tokens, wav_lens, tokens_lens, blank_index=self.cfg["blank_index"], reduction=self.cfg["loss_reduction"]).sum()
                loss_seq = self.criterion["seq_loss"](p_seq, tokens_eos, length=tokens_eos_lens, label_smoothing=self.cfg["label_smoothing"], reduction=self.cfg["loss_reduction"]).sum()
                loss_ctc = self.criterion["ctc_loss"](p_ctc, tokens, wav_lens, tokens_lens, blank_index=self.cfg["blank_index"], reduction=self.cfg["loss_reduction"]).sum()

                loss = (self.cfg["ctc_weight"] * loss_ctc + (1 - self.cfg["ctc_weight"]) * loss_seq)
                predicted_words = [self.tokenizer.decode_ids(utt_seq.tolist()).split(" ") for utt_seq in p_seq.argmax(-1)]
                target_words = [wrd.split(" ") for wrd in batch.wrd]
                self.metric.append(ids, predicted_words, target_words)

                eval_loss = loss.detach().cpu()
                avg_valid_loss = update_average(eval_loss, avg_valid_loss, step)

            wer = self.metric.summarize("error_rate")
            self.logger.info(f"epoch: {epoch}, time: {time.time()-eval_start_time}, wer: {wer}, avg_loss: {avg_valid_loss}")
        return wer


import torch
import logging
import functools
from asr.data.dataio.dataio import length_to_mask


logger = logging.getLogger(__name__)


def ctc_loss(
    log_probs, targets, input_lens, target_lens, blank_index, reduction="mean"
):
    """CTC loss.

    Arguments
    ---------
    predictions : torch.Tensor
        Predicted tensor, of shape [batch, time, chars].
    targets : torch.Tensor
        Target tensor, without any blanks, of shape [batch, target_len]
    input_lens : torch.Tensor
        Length of each utterance.
    target_lens : torch.Tensor
        Length of each target sequence.
    blank_index : int
        The location of the blank symbol among the character indexes.
    reduction : str
        What reduction to apply to the output. 'mean', 'sum', 'batch',
        'batchmean', 'none'.
        See pytorch for 'mean', 'sum', 'none'. The 'batch' option returns
        one loss per item in the batch, 'batchmean' returns sum / batch size.
    """
    input_lens = (input_lens * log_probs.shape[1]).round().int()
    target_lens = (target_lens * targets.shape[1]).round().int()
    log_probs = log_probs.transpose(0, 1)

    if reduction == "batchmean":
        reduction_loss = "sum"
    elif reduction == "batch":
        reduction_loss = "none"
    else:
        reduction_loss = reduction
    loss = torch.nn.functional.ctc_loss(
        log_probs,
        targets,
        input_lens,
        target_lens,
        blank_index,
        zero_infinity=True,
        reduction=reduction_loss,
    )

    if reduction == "batchmean":
        return loss / targets.shape[0]
    elif reduction == "batch":
        N = loss.size(0)
        return loss.view(N, -1).sum(1) / target_lens.view(N, -1).sum(1)
    else:
        return loss


def nll_loss(
    log_probabilities,
    targets,
    length=None,
    label_smoothing=0.0,
    allowed_len_diff=3,
    reduction="mean",
):
    """Computes negative log likelihood loss.

    Arguments
    ---------
    log_probabilities : torch.Tensor
        The probabilities after log has been applied.
        Format is [batch, log_p] or [batch, frames, log_p].
    targets : torch.Tensor
        The targets, of shape [batch] or [batch, frames].
    length : torch.Tensor
        Length of each utterance, if frame-level loss is desired.
    allowed_len_diff : int
        Length difference that will be tolerated before raising an exception.
    reduction : str
        Options are 'mean', 'batch', 'batchmean', 'sum'.
        See pytorch for 'mean', 'sum'. The 'batch' option returns
        one loss per item in the batch, 'batchmean' returns sum / batch size.

    Example
    -------
    >>> probs = torch.tensor([[0.9, 0.1], [0.1, 0.9]])
    >>> nll_loss(torch.log(probs), torch.tensor([1, 1]))
    tensor(1.2040)
    """
    if len(log_probabilities.shape) == 3:
        log_probabilities, targets = truncate(
            log_probabilities, targets, allowed_len_diff
        )
        log_probabilities = log_probabilities.transpose(1, -1)

    # Pass the loss function but apply reduction="none" first
    loss = functools.partial(torch.nn.functional.nll_loss, reduction="none")
    return compute_masked_loss(
        loss,
        log_probabilities,
        targets.long(),
        length,
        label_smoothing=label_smoothing,
        reduction=reduction,
    )


def kldiv_loss(
    log_probabilities,
    targets,
    length=None,
    label_smoothing=0.0,
    allowed_len_diff=3,
    pad_idx=0,
    reduction="mean",
):
    """Computes the KL-divergence error at the batch level.
    This loss applies label smoothing directly to the targets

    Arguments
    ---------
    probabilities : torch.Tensor
        The posterior probabilities of shape
        [batch, prob] or [batch, frames, prob].
    targets : torch.Tensor
        The targets, of shape [batch] or [batch, frames].
    length : torch.Tensor
        Length of each utterance, if frame-level loss is desired.
    allowed_len_diff : int
        Length difference that will be tolerated before raising an exception.
    reduction : str
        Options are 'mean', 'batch', 'batchmean', 'sum'.
        See pytorch for 'mean', 'sum'. The 'batch' option returns
        one loss per item in the batch, 'batchmean' returns sum / batch size.

    Example
    -------
    >>> probs = torch.tensor([[0.9, 0.1], [0.1, 0.9]])
    >>> kldiv_loss(torch.log(probs), torch.tensor([1, 1]))
    tensor(1.2040)
    """
    if label_smoothing > 0:
        if log_probabilities.dim() == 2:
            log_probabilities = log_probabilities.unsqueeze(1)

        bz, time, n_class = log_probabilities.shape
        targets = targets.long().detach()

        confidence = 1 - label_smoothing

        log_probabilities = log_probabilities.view(-1, n_class)
        targets = targets.view(-1)
        with torch.no_grad():
            true_distribution = log_probabilities.clone()
            true_distribution.fill_(label_smoothing / (n_class - 1))
            ignore = targets == pad_idx
            targets = targets.masked_fill(ignore, 0)
            true_distribution.scatter_(1, targets.unsqueeze(1), confidence)

        loss = torch.nn.functional.kl_div(
            log_probabilities, true_distribution, reduction="none"
        )
        loss = loss.masked_fill(ignore.unsqueeze(1), 0)

        # return loss according to reduction specified
        if reduction == "mean":
            return loss.sum().mean()
        elif reduction == "batchmean":
            return loss.sum() / bz
        elif reduction == "batch":
            return loss.view(bz, -1).sum(1) / length
        elif reduction == "sum":
            return loss.sum()
        else:
            return loss
    else:
        return nll_loss(log_probabilities, targets, length, reduction=reduction)


def truncate(predictions, targets, allowed_len_diff=3):
    """Ensure that predictions and targets are the same length.

    Arguments
    ---------
    predictions : torch.Tensor
        First tensor for checking length.
    targets : torch.Tensor
        Second tensor for checking length.
    allowed_len_diff : int
        Length difference that will be tolerated before raising an exception.
    """
    len_diff = predictions.shape[1] - targets.shape[1]
    if len_diff == 0:
        return predictions, targets
    elif abs(len_diff) > allowed_len_diff:
        raise ValueError(
            "Predictions and targets should be same length, but got %s and "
            "%s respectively." % (predictions.shape[1], targets.shape[1])
        )
    elif len_diff < 0:
        return predictions, targets[:, : predictions.shape[1]]
    else:
        return predictions[:, : targets.shape[1]], targets


def compute_masked_loss(
    loss_fn,
    predictions,
    targets,
    length=None,
    label_smoothing=0.0,
    reduction="mean",
):
    """Compute the true average loss of a set of waveforms of unequal length.

    Arguments
    ---------
    loss_fn : function
        A function for computing the loss taking just predictions and targets.
        Should return all the losses, not a reduction (e.g. reduction="none").
    predictions : torch.Tensor
        First argument to loss function.
    targets : torch.Tensor
        Second argument to loss function.
    length : torch.Tensor
        Length of each utterance to compute mask. If None, global average is
        computed and returned.
    label_smoothing: float
        The proportion of label smoothing. Should only be used for NLL loss.
        Ref: Regularizing Neural Networks by Penalizing Confident Output
        Distributions. https://arxiv.org/abs/1701.06548
    reduction : str
        One of 'mean', 'batch', 'batchmean', 'none' where 'mean' returns a
        single value and 'batch' returns one per item in the batch and
        'batchmean' is sum / batch_size and 'none' returns all.
    """
    mask = torch.ones_like(targets)
    if length is not None:
        length_mask = length_to_mask(
            length * targets.shape[1], max_len=targets.shape[1],
        )

        # Handle any dimensionality of input
        while len(length_mask.shape) < len(mask.shape):
            length_mask = length_mask.unsqueeze(-1)
        length_mask = length_mask.type(mask.dtype)
        mask *= length_mask

    # Compute, then reduce loss
    loss = loss_fn(predictions, targets) * mask
    N = loss.size(0)
    if reduction == "mean":
        loss = loss.sum() / torch.sum(mask)
    elif reduction == "batchmean":
        loss = loss.sum() / N
    elif reduction == "batch":
        loss = loss.reshape(N, -1).sum(1) / mask.reshape(N, -1).sum(1)

    if label_smoothing == 0:
        return loss
    else:
        loss_reg = torch.mean(predictions, dim=1) * mask
        if reduction == "mean":
            loss_reg = torch.sum(loss_reg) / torch.sum(mask)
        elif reduction == "batchmean":
            loss_reg = torch.sum(loss_reg) / targets.shape[0]
        elif reduction == "batch":
            loss_reg = loss_reg.sum(1) / mask.sum(1)

        return -label_smoothing * loss_reg + (1 - label_smoothing) * loss



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import copy

import torch
from torch import nn

from module.nlp.Linear_super import LinearSuper as SuperLinear
from module.nlp.layernorm_super import LayerNormSuper as SuperBertLayerNorm
from module.nlp.bert_attention_super import SuperBertAttention
from module.nlp.bert_intermediate_super import SuperBertIntermediate


logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
                    datefmt='%m/%d/%Y %H:%M:%S',
                    level=logging.INFO)
logger = logging.getLogger(__name__)

class SuperBertOutput(nn.Module):
    def __init__(self, config):
        super(SuperBertOutput, self).__init__()
        self.dense = SuperLinear(config.intermediate_size, config.hidden_size)
        self.LayerNorm = SuperBertLayerNorm(config.hidden_size)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def set_sample_config(self, intermediate_size, sample_embed_dim):
        self.dense.set_sample_config(intermediate_size, sample_embed_dim)
        self.LayerNorm.set_sample_config(sample_embed_dim)

    def calc_sampled_param_num(self):
        dense_numel = self.dense.calc_sampled_param_num()
        ln_numel = self.LayerNorm.calc_sampled_param_num()

        #logger.info('dense_numel: {}\n'.format(dense_numel))
        #logger.info('ln_numel: {}\n'.format(ln_numel))
        return dense_numel + ln_numel

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class SuperBertLayer(nn.Module):
    def __init__(self, config):
        super(SuperBertLayer, self).__init__()
        self.attention = SuperBertAttention(config)
        self.intermediate = SuperBertIntermediate(config)
        self.output = SuperBertOutput(config)

    def set_sample_config(self, sample_embed_dim, intermediate_size, num_attention_head, qkv_size):
        self.attention.set_sample_config(sample_embed_dim, num_attention_head, qkv_size)
        self.intermediate.set_sample_config(sample_embed_dim, intermediate_size)
        self.output.set_sample_config(intermediate_size, sample_embed_dim)

    def calc_sampled_param_num(self):
        attention_numel = self.attention.calc_sampled_param_num()
        intermediate_numel = self.intermediate.calc_sampled_param_num()
        output_numel = self.output.calc_sampled_param_num()

        #logger.info('attention_numel: {}\n'.format(attention_numel))
        #logger.info('intermediate_numel: {}\n'.format(intermediate_numel))
        #logger.info('output_numel: {}\n'.format(output_numel))

        return attention_numel + intermediate_numel + output_numel

    def forward(self, hidden_states, attention_mask):

        attention_output = self.attention(hidden_states, attention_mask)
        attention_output, layer_att = attention_output

        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output, layer_att


class SuperBertEncoder(nn.Module):
    def __init__(self, config):
        super(SuperBertEncoder, self).__init__()
        layer = SuperBertLayer(config)
        self.layers = nn.ModuleList([copy.deepcopy(layer) for _ in range(config.num_hidden_layers)])
        self.sample_layer_num = None

    def set_sample_config(self, subbert_config):
        self.sample_layer_num = subbert_config['sample_layer_num']
        self.sample_embed_dim = subbert_config['sample_hidden_size']
        self.num_attention_heads = subbert_config['sample_num_attention_heads']
        self.itermediate_sizes = subbert_config['sample_intermediate_sizes']
        self.qkv_sizes = subbert_config['sample_qkv_sizes']
        for layer, num_attention_head, intermediate_size, qkv_size in zip(self.layers[:self.sample_layer_num],
                                                                          self.num_attention_heads,
                                                                          self.itermediate_sizes,
                                                                          self.qkv_sizes):
            layer.set_sample_config(self.sample_embed_dim, intermediate_size, num_attention_head, qkv_size)

    def calc_sampled_param_num(self):
        layers_numel = 0

        for layer in self.layers[:self.sample_layer_num]:
            layers_numel += layer.calc_sampled_param_num()

        #logger.info('layer_numel: {}'.format(layers_numel))

        return layers_numel

    def forward(self, hidden_states, attention_mask):
        all_encoder_layers = []
        all_encoder_att = []

        for i, layer_module in enumerate(self.layers[:self.sample_layer_num]):
            all_encoder_layers.append(hidden_states)
            hidden_states = layer_module(all_encoder_layers[i], attention_mask)
            hidden_states, layer_att = hidden_states
            all_encoder_att.append(layer_att)

        all_encoder_layers.append(hidden_states)

        return all_encoder_layers, all_encoder_att




import torch.nn as nn


class LayerNormBase(nn.LayerNorm):
    def __init__(self, embed_dim, eps=1e-5, elementwise_affine=True):
        super().__init__(embed_dim, eps, elementwise_affine)

    def forward(self, x):
        raise NotImplementedError



import ast
import os
import logging
import torch

from e2eAIOK.common.trainer.model.model_builder_nlp import ModelBuilderNLP
from e2eAIOK.DeNas.nlp.supernet_bert import SuperBertForQuestionAnswering, BertConfig
from e2eAIOK.DeNas.nlp.utils import decode_arch

class ModelBuilderNLPDeNas(ModelBuilderNLP):
    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_model(self):
        config = BertConfig.from_pretrained(self.cfg.model, num_labels=self.cfg.num_labels, finetuning_task=self.cfg.task_name)
        model = SuperBertForQuestionAnswering.from_pretrained(self.cfg.model, config)
        device = torch.device("cuda" if torch.cuda.is_available() and not self.args.no_cuda else "cpu")
        model.to(device)
        subbert_config = decode_arch(self.cfg.best_model_structure)
        model.module.set_sample_config(subbert_config) if hasattr(model, 'module') else model.set_sample_config(subbert_config)
        size = model.module.calc_sampled_param_num() if hasattr(model, 'module') else model.calc_sampled_param_num()
        print("architecture: {}".format(subbert_config))
        print('Total parameters: {}'.format(size))
        return model


def init_settings():
    settings = {}
    # start db to store history queries / models and score board
    settings['server'] = None
    return settings


def init_meta():
    meta = {}
    meta['dataset_format'] = 'forward'
    return meta


def init_advisor_params():
    params = {}
    params['ppn'] = 1
    params['cores'] = None
    params['hosts'] = ['localhost']
    params['ccl_worker_num'] = 1
    params['python_executable'] = None
    params['global_batch_size'] = 1024
    params['num_epochs'] = 1
    params['model_dir'] = "./"
    params['observation_budget'] = 1
    params['save_path'] = '/home/vmagent/app/e2eaiok/result/'
    return params

def default_settings(model_name, settings):
    default = {"model_name": model_name, "enable_sigopt": False, "interative": False}
    for k, v in default.items():
        if k not in settings:
            settings[k] = v

    return settings



import subprocess
import yaml
import logging
import time
import os

from e2eAIOK.common.utils import *
from e2eAIOK.SDA.modeladvisor.BaseModelAdvisor import BaseModelAdvisor

class DLRMAdvisor(BaseModelAdvisor):
    '''
        Wide and Deep sigopt model optimization launcher
    '''
    def __init__(self, dataset_meta_path, train_path, eval_path, args):
        args = setting_local_default(args)
        super().__init__(dataset_meta_path, train_path, eval_path, args)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')
        self.train_python = "/opt/intel/oneapi/intelpython/latest/envs/pytorch_mlperf/bin/python"
        self.train_script = "/home/vmagent/app/e2eaiok/modelzoo/dlrm/dlrm/launch.py"
        self.train_path = train_path
        self.test_path = eval_path
        self.dataset_meta_path = dataset_meta_path

    def initialize_model_parameter(self, assignments=None):
        '''
            Add model specific parameters
            For sigopt parameters, set parameter explicitly and the parameter will not be optimized by sigopt
        '''
        config = {}
        tuned_parameters = {}
        if assignments:
            mlp_top_size = ["1024-1024-512-256-1","512-512-256-128-1","512-256-128-1","512-256-1","256-128-1","128-64-1","256-1","128-1"]
            mlp_bot_size = ["13-512-256-","13-512-256-","13-256-","13-128-"]
            tuned_parameters['mlp_top_size'] = assignments["mlp_top_size"]
            tuned_parameters['mlp_bot_size'] = assignments["mlp_bot_size"]
            tuned_parameters['lamb_lr'] = str(assignments['lamb_lr'])
            tuned_parameters['learning_rate'] = str(assignments['learning_rate'])
            tuned_parameters['lr_num_warmup_steps'] = str(assignments['lr_num_warmup_steps'])
            tuned_parameters['lr_decay_start_step'] = str(assignments['lr_decay_start_step'])
            tuned_parameters['lr_num_decay_steps'] = str(assignments['lr_num_decay_steps'])
            tuned_parameters['arch_sparse_feature_size'] = str(assignments["arch_sparse_feature_size"])
            tuned_parameters['arch_mlp_top'] = str(mlp_top_size[tuned_parameters["mlp_top_size"]])
            tuned_parameters['arch_mlp_bot'] = str(mlp_bot_size[tuned_parameters["mlp_bot_size"]])+str(assignments["arch_sparse_feature_size"])
        else:
            mlp_top_size = ["1024-1024-512-256-1","512-512-256-128-1","512-256-128-1","512-256-1","256-128-1","128-64-1","256-1","128-1"]
            mlp_bot_size = ["13-512-256-","13-512-256-","13-256-","13-128-"]
            tuned_parameters['mlp_top_size'] = 4
            tuned_parameters['mlp_bot_size'] = 3
            tuned_parameters['lamb_lr'] = "16"
            tuned_parameters['learning_rate'] = "16"
            tuned_parameters['lr_num_warmup_steps'] = "4000"
            tuned_parameters['lr_decay_start_step'] = "5760"
            tuned_parameters['lr_num_decay_steps'] = "17000"
            tuned_parameters['arch_sparse_feature_size'] = 128
            tuned_parameters['arch_mlp_top'] = str(mlp_top_size[tuned_parameters['mlp_top_size']])
            tuned_parameters['arch_mlp_bot'] = str(mlp_bot_size[tuned_parameters['mlp_bot_size']])+str(128)
        config['tuned_parameters'] = tuned_parameters
        self.params['model_parameter'] = config
        self.params['model_saved_path'] = os.path.join(
            self.params['save_path'], get_hash_string(str(tuned_parameters)))
        return config
    
    ###### Implementation of required methods ######

    def generate_sigopt_yaml(self, file='dlrm_sigopt.yaml'):
        config = {}
        config['project'] = 'dlrm'
        config['experiment'] = 'DLRM'
        parameters = [{'name':'learning_rate','bounds':{'min':5,'max':50},'type':'int'},
                      {'name':'lamb_lr','bounds':{'min':5,'max':50},'type':'int'},
                      {'name':'lr_num_warmup_steps','bounds':{'min':2000,'max':4500},'type':'int'},
                      {'name':'lr_decay_start_step','bounds':{'min':4501,'max':9000},'type':'int'},
                      {'name':'lr_num_decay_steps','bounds':{'min':5000,'max':15000},'type':'int'},
                      {'name':'arch_sparse_feature_size','grid': [128,64,16],'type':'int'},
                      {'name':'mlp_top_size','bounds':{'min':0,'max':7},'type':'int'},
                      {'name':'mlp_bot_size','bounds':{'min':0,'max':3},'type':'int'},]
        config['parameters'] = parameters        
        metrics = []
        if 'training_time_threshold' in self.params:
            metrics.append({'name': 'training_time', 'objective': 'minimize', 'threshold': self.params['training_time_threshold']})
            metrics.append({'name': self.params['metric'], 'objective': self.params['metric_objective'], 'threshold': self.params['metric_threshold']})
        else:
            metrics.append({'name': self.params['metric'], 'objective': self.params['metric_objective']})
        
        config['metrics'] = metrics
        config['observation_budget'] = self.params['observation_budget']


        with open(file, 'w') as f:
            yaml.dump(config, f)
        return config

    def update_metrics(self):
        metrics = []
        metrics.append({'name': 'accuracy', 'value': self.mean_accuracy})
        metrics.append({'name': 'training_time', 'value': self.training_time})
        self.params['model_parameter']['metrics'] = metrics
        return self.params['model_parameter']['metrics']
  
    def train_model(self, args):
        start_time = time.time()
        if args['ppn'] == 1 and len(args['hosts']) == 1:
            self.launch(args)
        else:
            self.dist_launch(args)
        self.training_time = time.time() - start_time
        with open("./best_auc.txt",'r') as f:
            lines = f.readlines()
        self.mean_accuracy = float(lines[-1])
        metrics = self.update_metrics()
        model_path = args['model_saved_path']
        return self.training_time, model_path, metrics

    def launch(self, args):
        # construct WnD launch command
        cmd = f"{self.train_python} -u {self.train_script} "
        model_saved_path = args['model_saved_path']
        cmd +=f"/home/vmagent/app/e2eaiok/modelzoo/dlrm/dlrm/dlrm_s_pytorch.py --mini-batch-size={args['train_batch_size']} --print-freq=16  " \
            + f"--test-mini-batch-size={args['test_batch_size']} --test-freq=800 " \
            + f"--train-data-path={self.train_path} --eval-data-path={self.test_path} " \
            + f"--nepochs=1 --day-feature-count={args['data_path'] + '/day_fea_count.npz'} " \
            + f"--loss-function=bce --round-targets=True --num-workers=0 --test-num-workers=0 --use-ipex " \
            + f" --dist-backend=ccl --print-time --data-generation=dataset --optimizer=1 --bf16 --data-set=terabyte " \
            + f"--sparse-dense-boundary=403346  --memory-map --mlperf-logging --mlperf-auc-threshold=0.8025 " \
            + f"--mlperf-bin-loader --mlperf-bin-shuffle --numpy-rand-seed=12345 " \
            + f"--arch-sparse-feature-size={args['model_parameter']['tuned_parameters']['arch_sparse_feature_size']} " \
            + f"--arch-mlp-bot={args['model_parameter']['tuned_parameters']['arch_mlp_bot']} " \
            + f"--arch-mlp-top={args['model_parameter']['tuned_parameters']['arch_mlp_top']} " \
            + f"--lamblr={args['model_parameter']['tuned_parameters']['lamb_lr']} --save-model={model_saved_path} " \
            + f"--learning-rate={args['model_parameter']['tuned_parameters']['learning_rate']} " \
            + f"--lr-num-warmup-steps={args['model_parameter']['tuned_parameters']['lr_num_warmup_steps']} " \
            + f"--lr-decay-start-step={args['model_parameter']['tuned_parameters']['lr_decay_start_step']} " \
            + f"--lr-num-decay-steps={args['model_parameter']['tuned_parameters']['lr_num_decay_steps']}"
        self.logger.info(f'### Starting model training ###, launch cmd is: \n{cmd}')
        process = subprocess.Popen(cmd, shell=True)
        process.wait()
    
    def dist_launch(self, args):
        ppn = args['ppn']
        hosts = args['hosts']
        model_saved_path = args['model_saved_path']
        if len(args['hosts']) == 1:
            cmd = f"{self.train_python} -u {self.train_script} --distributed --nproc_per_node={ppn} --nnodes={len(hosts)} --hostfile {','.join(hosts)} "
        else:
            hostfile = args['hostfile']
            cmd = f"{self.train_python} -u {self.train_script}  --distributed --nproc_per_node={ppn} --nnodes={len(hosts)} --hostfile {hostfile} "  
        cmd +=f"/home/vmagent/app/e2eaiok/modelzoo/dlrm/dlrm/dlrm_s_pytorch.py --mini-batch-size={args['train_batch_size']} --print-freq=16  " \
            + f"--test-mini-batch-size={args['test_batch_size']} --test-freq=800 " \
            + f"--train-data-path={self.train_path} --eval-data-path={self.test_path} " \
            + f"--nepochs=1 --day-feature-count={args['data_path'] + '/day_fea_count.npz'} " \
            + f"--loss-function=bce --round-targets=True --num-workers=0 --test-num-workers=0 --use-ipex " \
            + f" --dist-backend=ccl --print-time --data-generation=dataset --optimizer=1 --bf16 --data-set=terabyte " \
            + f"--sparse-dense-boundary=403346  --memory-map --mlperf-logging --mlperf-auc-threshold=0.8025 " \
            + f"--mlperf-bin-loader --mlperf-bin-shuffle --numpy-rand-seed=12345 " \
            + f"--arch-sparse-feature-size={args['model_parameter']['tuned_parameters']['arch_sparse_feature_size']} " \
            + f"--arch-mlp-bot={args['model_parameter']['tuned_parameters']['arch_mlp_bot']} " \
            + f"--arch-mlp-top={args['model_parameter']['tuned_parameters']['arch_mlp_top']} " \
            + f"--lamblr={args['model_parameter']['tuned_parameters']['lamb_lr']} --save-model={model_saved_path} " \
            + f"--learning-rate={args['model_parameter']['tuned_parameters']['learning_rate']} " \
            + f"--lr-num-warmup-steps={args['model_parameter']['tuned_parameters']['lr_num_warmup_steps']} " \
            + f"--lr-decay-start-step={args['model_parameter']['tuned_parameters']['lr_decay_start_step']} " \
            + f"--lr-num-decay-steps={args['model_parameter']['tuned_parameters']['lr_num_decay_steps']}"
        self.logger.info(f'### Starting model training ###, launch cmd is: \n{cmd}')
        process = subprocess.Popen(cmd, shell=True)
        process.wait() 

def setting_local_default(settings):
    default = {
        "ppn": 2,
        "ccl_worker_num": 4,
        "train_batch_size": 262144,
        "test_batch_size": 131072,
        "num_epochs": 1,
        "cores": None,
        "training_time_threshold": 10000,
        "metric": "accuracy",
        "metric_objective": "maximize",
        "metric_threshold": 0.6,
    }
    for k, v in default.items():
        if k not in settings:
            settings[k] = v

    return settings



import logging
import subprocess
import yaml

from e2eAIOK.common.utils import *
from e2eAIOK.SDA.modeladvisor.BaseModelAdvisor import BaseModelAdvisor

class DIENAdvisor(BaseModelAdvisor):
    def __init__(self, dataset_meta_path, train_path, eval_path, settings):
        super().__init__(dataset_meta_path, train_path, eval_path, settings)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')
        self.total_cores, self.physical_cores = self.get_cpu_info()

        # set default required arguments
        if 'ppn' not in self.params:
            self.params['ppn'] = 1
        self.params['num_instances'] = self.physical_cores
        self.params['num_cores'] = int(self.total_cores / self.physical_cores)

        # check distributed configuration
        missing_params = []
        # mpirun -n 1 -hosts 172.16.8.30 -ppn 1 -iface ens21f1
        if self.params["ppn"] > 1:
            missing_params = missing_params + \
                ['hosts'] if 'hosts' not in self.params else missing_params
            missing_params = missing_params + \
                ['iface'] if 'iface' not in self.params else missing_params
        if len(missing_params) > 0:
            raise ValueError(
                f"[CONFIG ERROR] Missing parameters {missing_params} in \
                e2eaiok_defaults.conf when ppn is set above 1.")

        self.train_path = train_path
        self.test_path = eval_path
        self.dataset_meta_path = dataset_meta_path
        # self.saved_path = settings['model_saved_path']
        self.python_path = "/opt/intel/oneapi/intelpython/latest/envs/tensorflow/bin/"
        self.train_python = f"{self.python_path}/python"
        self.horovodrun_path = f"{self.python_path}/horovodrun"
        self.train_script = "/home/vmagent/app/e2eaiok/modelzoo/dien/train/ai-matrix/script/train.py"

    def get_cpu_info(self):
        # get cpu physical cores and virtual cores per core as return
        num_total_cores = 128
        num_physical_cores = 32
        return num_total_cores, num_physical_cores

    # ====== Implementation of required methods ======

    def update_metrics(self):
        result_metrics_path = os.path.join(self.params['model_saved_path'],
                                           "result.yaml")
        if not os.path.exists(result_metrics_path):
            raise FileNotFoundError(
                f"{self.train_script} completed, while we can't find \
                    result {result_metrics_path} file.")
        with open(result_metrics_path) as f:
            results = yaml.load(f, Loader=yaml.FullLoader)
        self.training_time = results['training_time']
        self.best_trained_model_path = results['best_trained_model']
        metrics = []
        metrics.append({'name': 'AUC', 'value': results['AUC']})
        metrics.append({
            'name': 'training_time',
            'value': results['training_time']
        })
        self.params['model_parameter']['metrics'] = metrics
        return self.params['model_parameter']['metrics']

    def initialize_model_parameter(self, assignments=None):
        config = {}
        tuned_parameters = {}
        if assignments:
            tuned_parameters['batch_size'] = assignments['batch_size']
        else:
            tuned_parameters['batch_size'] = 256
        config['tuned_parameters'] = tuned_parameters
        self.params['model_parameter'] = config
        self.params['model_saved_path'] = os.path.join(
            self.params['save_path'], get_hash_string(str(tuned_parameters)))
        return config

    def generate_sigopt_yaml(self, file='test_sigopt.yaml'):
        config = {}
        config['project'] = 'e2eaiok'
        config['experiment'] = 'dien'
        parameters = [{
            'name': 'batch_size',
            'categorical_values': ["256", "512", "1024"],
            'type': 'categorical'
        }]
        user_defined_parameter = self.params['model_parameter'][
            'parameters'] if ('model_parameter' in self.params) and (
                'parameters' in self.params['model_parameter']) else None
        config['parameters'] = parameters
        if user_defined_parameter:
            self.logger.info(
                f"Update with user defined parameters {user_defined_parameter}"
            )
            update_list(config['parameters'], user_defined_parameter)
        config['metrics'] = [{
            'name': 'AUC',
            'strategy': 'optimize',
            'objective': 'maximize'
        }, {
            'name': 'training_time',
            'objective': 'minimize'
        }]
        user_defined_metrics = self.params['model_parameter']['metrics'] if (
            'model_parameter' in self.params) and (
                'metrics' in self.params['model_parameter']) else None
        if user_defined_metrics:
            self.logger.info(
                f"Update with user defined parameters {user_defined_metrics}")
            update_list(config['metrics'], user_defined_metrics)
        config['observation_budget'] = self.params['observation_budget']

        # TODO: Add all parameter tuning here

        saved_path = os.path.join(self.params['save_path'], file)
        with open(saved_path, 'w') as f:
            yaml.dump(config, f)
        return config

    def train_model(self, args):
        if args['ppn'] > 1:
            self.dist_launch(args)
        else:
            self.launch(args)
        metrics = self.update_metrics()
        return self.training_time, self.best_trained_model_path, metrics

    def dist_launch(self, args):
        cmd = []
        hosts = [f"{h}:1" for h in args['hosts']]
        cmd.extend([self.horovodrun_path, "-np", f"{args['ppn']}", "-H",  f"{','.join(hosts)}", "--network-interface", f"{args['iface']}"])
        cmd.extend(["--verbose"])
        cmd.extend(self.prepare_cmd(args))

        self.logger.info(f'training launch command: {" ".join(cmd)}')
        process = subprocess.Popen(cmd)
        process.wait()

    def launch(self, args):
        cmd = []
        cmd.extend(self.prepare_cmd(args))
        self.logger.info(f'training launch command: {" ".join(cmd)}')
        process = subprocess.Popen(cmd)
        process.wait()

    def prepare_cmd(self, args):
        cmd = []
        cmd.extend([
            f"{self.train_python}", "-u", f"{self.train_script}", "--train_path",
            f"{self.train_path}", "--test_path", f"{self.test_path}",
            "--meta_path", f"{self.dataset_meta_path}"
        ])
        cmd.extend(["--saved_path", f"{args['model_saved_path']}"])
        cmd.extend(["--num-intra-threads", f"{args['num_instances']}"])
        cmd.extend(["--num-inter-threads", f"{args['num_cores']}"])

        # fixed parameters
        cmd.extend([
            "--mode", "train", "--embedding_device", "cpu", "--model", "DIEN"
        ])
        cmd.extend(["--slice_id", "0", "--advanced", "true", "--seed", "3"])
        cmd.extend(["--data_type", "FP32"])

        # tunnable parameters
        tuned_parameters = args['model_parameter']['tuned_parameters']
        cmd.extend(["--batch_size", f"{tuned_parameters['batch_size']}"])

        # fake result
        # result_metrics_path = os.path.join(self.params['model_saved_path'], "result.yaml")
        # result = {"AUC": 0.823, "training_time": 425, "best_trained_model": f"{self.params['model_saved_path']}/best_trained_model"}
        # with open(result_metrics_path, "w") as f:
        #    results = yaml.dump(result, f)
        return cmd



import torch
from e2eAIOK.common.trainer.data.data_builder_asr import DataBuilderASR
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist
from e2eAIOK.DeNas.asr.data.dataio.dataset import dataio_prepare
from e2eAIOK.DeNas.asr.data.dataio.batch import PaddedBatch

class DataBuilderLibriSpeech(DataBuilderASR):
    def __init__(self, cfg, tokenizer):
        super().__init__(cfg, tokenizer)

    def prepare_dataset(self):
        train_data, valid_data, test_datasets = dataio_prepare(self.cfg, self.tokenizer)
        self.dataset_train = train_data
        self.dataset_val = valid_data


import subprocess
import yaml
import logging
import time
import sys

from e2eAIOK.common.utils import *
from e2eAIOK.SDA.modeladvisor.BaseModelAdvisor import BaseModelAdvisor


class ResNetAdvisor(BaseModelAdvisor):
    def __init__(self, dataset_meta_path, train_path, eval_path, settings):
        super().__init__(dataset_meta_path, train_path, eval_path, settings)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')
        self.train_script = "/home/vmagent/app/e2eaiok/modelzoo/resnet/mlperf_resnet/imagenet_main.py"
        self.python_path = "/opt/intel/oneapi/intelpython/latest/envs/tensorflow/bin/"
        self.train_python = f"{self.python_path}/python"
        self.horovodrun_path = f"{self.python_path}/horovodrun"


    def update_metrics(self):
        metrics = []
        metrics.append({'name': 'accuracy', 'value': self.mean_accuracy})
        metrics.append({'name': 'training_time', 'value': self.training_time})
        self.params['model_parameter']['metrics'] = metrics
        return self.params['model_parameter']['metrics']

    def initialize_model_parameter(self, assignments=None):
        config = {}
        tuned_parameters = {}
        if assignments:
            tuned_parameters['label_smoothing'] = assignments['label_smoothing']
            tuned_parameters['resnet_size'] = assignments['resnet_size']
            tuned_parameters['num_filters'] = assignments['num_filters']
            tuned_parameters['kernel_size'] = assignments['kernel_size']
            tuned_parameters['momentum'] = assignments['momentum']
            tuned_parameters['weight_decay'] = assignments['weight_decay']
            tuned_parameters['base_lr'] = assignments['base_lr']
        else:
            tuned_parameters['label_smoothing'] = 0.0
            tuned_parameters['resnet_size'] = 50
            tuned_parameters['num_filters'] = 64
            tuned_parameters['kernel_size'] = 7
            tuned_parameters['momentum'] = 0.9
            tuned_parameters['weight_decay'] = 1e-4
            tuned_parameters['base_lr'] = 0.128
        config['tuned_parameters'] = tuned_parameters
        self.params['model_parameter'] = config
        self.params['model_saved_path'] = os.path.join(
            self.params['save_path'], get_hash_string(str(tuned_parameters)))
        return config

    def generate_sigopt_yaml(self, file='resnet_sigopt.yaml'):
        config = {}
        config['project'] = self.params['project']
        config['experiment'] = self.params['experiment']
        parameters = [{'name': 'label_smoothing', 'bounds': {'min': 0.05, 'max': 0.5}, 'type': 'double'},
                      {'name': 'resnet_size', 'grid': [18, 34, 50], 'type': 'int'},
                      {'name': 'num_filters', 'grid': [2,4,8,16,32,64], 'type': 'int'},
                      {'name': 'kernel_size', 'grid': [1,3,5,7,9], 'type': 'int'},
                      {'name': 'momentum', 'bounds': {'min': 0.6, 'max': 0.99}, 'type': 'double'},
                      {'name': 'weight_decay', 'bounds': {'min': 1e-5, 'max': 5e-5}, 'type': 'double'},
                      {'name': 'base_lr', 'bounds': {'min': 0.01, 'max': 0.2}, 'type': 'double'},
                      ]
                      
        config['parameters'] = parameters
        metrics = []
        if 'training_time_threshold' in self.params:
            metrics.append({'name': 'training_time', 'objective': 'minimize', 'threshold': self.params['training_time_threshold']})
            metrics.append({'name': self.params['metric'], 'objective': self.params['metric_objective'], 'threshold': self.params['metric_threshold']})
        else:
            metrics.append({'name': self.params['metric'], 'objective': self.params['metric_objective']})
        
        config['metrics'] = metrics
        config['observation_budget'] = self.params['observation_budget']
        with open(file, 'w') as f:
            yaml.dump(config, f)
        return config

    def train_model(self, args):
        start_time = time.time()
        if args['ppn'] == 1 and len(args['hosts']) == 1:
            self.dist_launch(args)
        else:
            self.dist_launch(args)
        self.training_time = time.time() - start_time
        with open("/home/vmagent/app/e2eaiok/modelzoo/resnet/mlperf_resnet/metric.txt",'r') as f:
            lines = f.readlines()
        self.mean_accuracy = float(lines[-1])
        metrics = self.update_metrics()
        model_path = self.model_saved_path
        return self.training_time, model_path, metrics

    def launch(self, args):
        # construct ResNet launch command
        self.model_saved_path = args['model_saved_path']
        cmd = f"{self.train_python} -u {self.train_script} '123456' "
        cmd += f"--label_smoothing '{args['model_parameter']['tuned_parameters']['label_smoothing']}' --num_filters '{args['model_parameter']['tuned_parameters']['num_filters']}' --data_dir '{args['train_dataset_path']}' " \
            + f"--model_dir '{self.model_saved_path}' --train_epochs '{args['num_epochs']}' " \
            + f"--stop_threshold '{args['metric_threshold']}' --batch_size '{args['global_batch_size']}' --version 1 --resnet_size '{args['model_parameter']['tuned_parameters']['resnet_size']}' " \
            + f"--epochs_between_evals 4 --inter_op_parallelism_threads 2 --intra_op_parallelism_threads 2 " \
            + f"--enable_lars --weight_decay '{args['model_parameter']['tuned_parameters']['weight_decay']}' --kernel_size '{args['model_parameter']['tuned_parameters']['kernel_size']}'"
        if args["use_synthetic_data"]:
            cmd += f"--use_synthetic_data"
        
        print(f'training launch command: {cmd}')
        process = subprocess.Popen(cmd, shell=True)
        process.wait() 

    def dist_launch(self, args):
        # construct ResNet distributed launch command
        ppn = args['ppn']
        hosts = ",".join(str(i)+":"+str(ppn) for i in args['hosts']) 
        self.model_saved_path = args['model_saved_path']
        ranks = len(args['hosts']) * ppn
        cmd = f"{self.horovodrun_path} -n {ranks} -H {hosts} HOROVOD_CPU_OPERATIONS=CCL CCL_ATL_TRANSPORT=mpi "
        cmd += f"{self.train_python} -u {self.train_script} '123456' "
        cmd += f"--label_smoothing '{args['model_parameter']['tuned_parameters']['label_smoothing']}' --num_filters '{args['model_parameter']['tuned_parameters']['num_filters']}' --data_dir '{args['train_dataset_path']}' " \
            + f"--model_dir '{self.model_saved_path}' --train_epochs '{args['num_epochs']}' " \
            + f"--stop_threshold '{args['metric_threshold']}' --batch_size '{204}' --version 1 --resnet_size '{args['model_parameter']['tuned_parameters']['resnet_size']}' " \
            + f"--epochs_between_evals 1 --inter_op_parallelism_threads 2 --intra_op_parallelism_threads 2 " \
            + f"--enable_lars --weight_decay '{args['model_parameter']['tuned_parameters']['weight_decay']}' --kernel_size '{args['model_parameter']['tuned_parameters']['kernel_size']}' " \
            + f"--base_lr '{args['model_parameter']['tuned_parameters']['base_lr']}' --momentum '{args['model_parameter']['tuned_parameters']['momentum']}' "

        if args["use_synthetic_data"]:
            cmd += f"--use_synthetic_data"

        self.logger.info(f'training launch command: {cmd}')
        print(f'training launch command: {cmd}')
        process = subprocess.Popen(cmd, shell=True)
        process.wait() 



import sys
import torch
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist

class DataBuilder():
    """
    The basic data builder class for all dataset

    Note:
        You should implement specfic build_dataset function in the build_dataset.py under data folder
    """
    def __init__(self,cfg):
        self.cfg = cfg
        self.dataset_train = None
        self.dataset_val = None
        self.dataset_test = None
    
    def prepare_dataset(self):
        """
            prepare training/evaluation dataset
        """
        raise NotImplementedError("prepare_dataset is abstract.")
    
    def get_dataloader(self):
        """
            create training/evaluation dataloader
        """
        if self.dataset_train is None or self.dataset_val is None:
            self.prepare_dataset()

        if ext_dist.my_size > 1:
            num_tasks = ext_dist.dist.get_world_size()
            global_rank = ext_dist.dist.get_rank()
            sampler_train = torch.utils.data.DistributedSampler(
                self.dataset_train, num_replicas=num_tasks, rank=global_rank
            )
            
            sampler_val = torch.utils.data.DistributedSampler(
                self.dataset_val, num_replicas=num_tasks, rank=global_rank)
        else:
            sampler_val = None
            sampler_train = None

        dataloader_train = torch.utils.data.DataLoader(
            self.dataset_train, 
            sampler=sampler_train,
            batch_size=self.cfg.train_batch_size,
            num_workers=self.cfg.num_workers,
            shuffle=True,
            drop_last=True,
        )

        dataloader_val = torch.utils.data.DataLoader(
            self.dataset_val, 
            batch_size=self.cfg.eval_batch_size,
            sampler=sampler_val, 
            num_workers=self.cfg.num_workers,
            shuffle=False,
            drop_last=False
        )
        
        if self.dataset_test is None:
            return dataloader_train, dataloader_val
        else:
            dataloader_test = torch.utils.data.DataLoader(
            self.dataset_test, 
            batch_size=self.cfg.eval_batch_size,
            sampler=sampler_val, 
            num_workers=self.cfg.num_workers,
            shuffle=False,
            drop_last=False
        )
        return dataloader_train, dataloader_val, dataloader_test
